from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
)

from uuid import UUID
from threading import Thread
from rich import print
from uuid import uuid4

from litellm.utils import ModelResponse, get_max_tokens
from promptmodel.llms.llm import LLM
from promptmodel.database.models import (
    DeployedPrompt,
    DeployedFunctionModel,
    DeployedFunctionModelVersion,
)
from promptmodel.database.crud import (
    get_deployed_prompts,
)
from promptmodel.promptmodel_init import CacheManager
from promptmodel.utils.config_utils import read_config, upsert_config
from promptmodel.utils.random_utils import select_version_by_ratio
from promptmodel.utils import logger
from promptmodel.utils.async_utils import run_async_in_sync
from promptmodel.utils.token_counting import (
    num_tokens_for_messages_for_each,
    num_tokens_from_functions_input,
)
from promptmodel.utils.output_utils import update_dict
from promptmodel.apis.base import AsyncAPIClient
from promptmodel.types.response import (
    LLMResponse,
    LLMStreamResponse,
    FunctionModelConfig,
    ChatModelConfig,
    PMDetail,
)
from promptmodel.types.request import ChatLogRequest


class LLMProxy(LLM):
    def __init__(self, name: str, version: Optional[Union[str, int]] = "deploy"):
        super().__init__()
        self._name = name
        self.version = version

    def _wrap_gen(self, gen: Callable[..., Any]) -> Callable[..., Any]:
        def wrapper(inputs: Dict[str, Any], **kwargs):
            prompts, version_details = run_async_in_sync(
                LLMProxy.fetch_prompts(self._name, self.version)
            )
            call_args = self._prepare_call_args(
                prompts, version_details, inputs, kwargs
            )

            log_uuid = str(uuid4())

            # Call the generator with the arguments
            stream_response: Generator[LLMStreamResponse, None, None] = gen(**call_args)

            api_response = None
            dict_cache = {}  # to store aggregated dictionary values
            string_cache = ""  # to store aggregated string values
            error_occurs = False
            error_log = None
            for item in stream_response:
                if (
                    item.api_response and "delta" not in item.api_response.choices[0]
                ):  # only get the last api_response, not delta response
                    api_response = item.api_response
                if item.parsed_outputs:
                    dict_cache = update_dict(dict_cache, item.parsed_outputs)
                if item.raw_output:
                    string_cache += item.raw_output
                if item.error and not error_occurs:
                    error_occurs = True
                    error_log = item.error_log

                if error_occurs:
                    # delete all promptmodel data in item
                    item.raw_output = None
                    item.parsed_outputs = None
                    item.function_call = None
                item.pm_detail = PMDetail(
                    model=version_details["model"],
                    name=self._name,
                    version_uuid=str(version_details["uuid"]),
                    version=version_details["version"],
                    log_uuid=log_uuid,
                )
                yield item

            metadata = {
                "error": error_occurs,
                "error_log": error_log,
            }

            run_async_in_sync(
                self._async_log_to_cloud(
                    log_uuid=log_uuid,
                    version_uuid=version_details["uuid"],
                    inputs=inputs,
                    api_response=api_response,
                    parsed_outputs=dict_cache,
                    metadata=metadata,
                )
            )

        return wrapper

    def _wrap_async_gen(self, async_gen: Callable[..., Any]) -> Callable[..., Any]:
        async def wrapper(inputs: Dict[str, Any], **kwargs):
            prompts, version_details = await LLMProxy.fetch_prompts(
                self._name, self.version
            )
            call_args = self._prepare_call_args(
                prompts, version_details, inputs, kwargs
            )

            # Call async_gen with the arguments
            stream_response: AsyncGenerator[LLMStreamResponse, None] = async_gen(
                **call_args
            )

            log_uuid = str(uuid4())

            api_response = None
            dict_cache = {}  # to store aggregated dictionary values
            string_cache = ""  # to store aggregated string values
            error_occurs = False
            error_log = None
            api_response: Optional[ModelResponse] = None
            async for item in stream_response:
                if (
                    item.api_response and "delta" not in item.api_response.choices[0]
                ):  # only get the last api_response, not delta response
                    api_response = item.api_response
                if item.parsed_outputs:
                    dict_cache = update_dict(dict_cache, item.parsed_outputs)
                if item.raw_output:
                    string_cache += item.raw_output
                if item.error and not error_occurs:
                    error_occurs = True
                    error_log = item.error_log
                item.pm_detail = PMDetail(
                    model=version_details["model"],
                    name=self._name,
                    version_uuid=str(version_details["uuid"]),
                    version=version_details["version"],
                    log_uuid=log_uuid,
                )
                yield item

            # # add string_cache in model_response
            # if api_response:
            #     if "message" not in api_response.choices[0]:
            #         api_response.choices[0].message = {}
            #     if "content" not in api_response.choices[0].message:
            #         api_response.choices[0].message["content"] = string_cache
            #         api_response.choices[0].message["role"] = "assistant"

            metadata = {
                "error": error_occurs,
                "error_log": error_log,
            }
            await self._async_log_to_cloud(
                log_uuid=log_uuid,
                version_uuid=version_details["uuid"],
                inputs=inputs,
                api_response=api_response,
                parsed_outputs=dict_cache,
                metadata=metadata,
            )

            # raise Exception("error_log")

        return wrapper

    def _wrap_method(self, method: Callable[..., Any]) -> Callable[..., Any]:
        def wrapper(inputs: Dict[str, Any], **kwargs):
            prompts, version_details = run_async_in_sync(
                LLMProxy.fetch_prompts(self._name, self.version)
            )
            call_args = self._prepare_call_args(
                prompts, version_details, inputs, kwargs
            )

            # Call the method with the arguments
            llm_response: LLMResponse = method(**call_args)
            error_occurs = llm_response.error
            error_log = llm_response.error_log
            metadata = {
                "error": error_occurs,
                "error_log": error_log,
            }
            log_uuid = str(uuid4())
            if llm_response.parsed_outputs:
                run_async_in_sync(
                    self._async_log_to_cloud(
                        log_uuid=log_uuid,
                        version_uuid=version_details["uuid"],
                        inputs=inputs,
                        api_response=llm_response.api_response,
                        parsed_outputs=llm_response.parsed_outputs,
                        metadata=metadata,
                    )
                )
            else:
                run_async_in_sync(
                    self._async_log_to_cloud(
                        log_uuid=log_uuid,
                        version_uuid=version_details["uuid"],
                        inputs=inputs,
                        api_response=llm_response.api_response,
                        parsed_outputs={},
                        metadata=metadata,
                    )
                )
            if error_occurs:
                # delete all promptmodel data in llm_response
                llm_response.raw_output = None
                llm_response.parsed_outputs = None
                llm_response.function_call = None

            llm_response.pm_detail = PMDetail(
                model=version_details["model"],
                name=self._name,
                version_uuid=str(version_details["uuid"]),
                version=version_details["version"],
                log_uuid=log_uuid,
            )
            return llm_response

        return wrapper

    def _wrap_async_method(self, method: Callable[..., Any]) -> Callable[..., Any]:
        async def async_wrapper(inputs: Dict[str, Any], **kwargs):
            prompts, version_details = await LLMProxy.fetch_prompts(
                self._name, self.version
            )  # messages, model, uuid = self._fetch_prompts()
            call_args = self._prepare_call_args(
                prompts, version_details, inputs, kwargs
            )

            # Call the method with the arguments
            llm_response: LLMResponse = await method(**call_args)
            error_occurs = llm_response.error
            error_log = llm_response.error_log
            metadata = {
                "error": error_occurs,
                "error_log": error_log,
            }
            log_uuid = str(uuid4())
            if llm_response.parsed_outputs:
                await self._async_log_to_cloud(
                    log_uuid=log_uuid,
                    version_uuid=version_details["uuid"],
                    inputs=inputs,
                    api_response=llm_response.api_response,
                    parsed_outputs=llm_response.parsed_outputs,
                    metadata=metadata,
                )
            else:
                await self._async_log_to_cloud(
                    log_uuid=log_uuid,
                    version_uuid=version_details["uuid"],
                    inputs=inputs,
                    api_response=llm_response.api_response,
                    parsed_outputs={},
                    metadata=metadata,
                )

            if error_occurs:
                # delete all promptmodel data in llm_response
                llm_response.raw_output = None
                llm_response.parsed_outputs = None
                llm_response.function_call = None

            llm_response.pm_detail = PMDetail(
                model=version_details["model"],
                name=self._name,
                version_uuid=str(version_details["uuid"]),
                version=version_details["version"],
                log_uuid=log_uuid,
            )
            return llm_response

        return async_wrapper

    def _wrap_chat(self, method: Callable[..., Any]) -> Callable[..., Any]:
        def wrapper(session_uuid: str, **kwargs):
            instruction, version_details, message_logs = run_async_in_sync(
                LLMProxy.fetch_chat_model(self._name, session_uuid, self.version)
            )

            call_args = self._prepare_call_args_for_chat(
                message_logs, version_details, kwargs
            )

            # Call the method with the arguments
            llm_response: LLMResponse = method(**call_args)
            error_occurs = llm_response.error
            error_log = llm_response.error_log
            metadata = {
                "error": error_occurs,
                "error_log": error_log,
            }
            api_response = None
            if llm_response.api_response:
                api_response = llm_response.api_response

            log_uuid = str(uuid4())

            run_async_in_sync(
                self._async_chat_log_to_cloud(
                    session_uuid=session_uuid,
                    version_uuid=version_details["uuid"],
                    chat_log_request_list=[
                        ChatLogRequest(
                            message=llm_response.api_response.choices[
                                0
                            ].message.model_dump(),
                            uuid=log_uuid,
                            metadata=metadata,
                            api_response=api_response,
                        )
                    ],
                )
            )

            if error_occurs:
                # delete all promptmodel data in llm_response
                llm_response.raw_output = None
                llm_response.parsed_outputs = None
                llm_response.function_call = None

            llm_response.pm_detail = PMDetail(
                model=version_details["model"],
                name=self._name,
                version_uuid=str(version_details["uuid"]),
                version=version_details["version"],
                log_uuid=log_uuid,
            )
            return llm_response

        return wrapper

    def _wrap_async_chat(self, method: Callable[..., Any]) -> Callable[..., Any]:
        async def async_wrapper(session_uuid: str, **kwargs):
            (
                instruction,
                version_details,
                message_logs,
            ) = await LLMProxy.fetch_chat_model(self._name, session_uuid, self.version)

            call_args = self._prepare_call_args_for_chat(
                message_logs, version_details, kwargs
            )

            # Call the method with the arguments
            llm_response: LLMResponse = await method(**call_args)
            error_occurs = llm_response.error
            error_log = llm_response.error_log
            metadata = {
                "error": error_occurs,
                "error_log": error_log,
            }
            api_response = None
            if llm_response.api_response:
                api_response = llm_response.api_response

            log_uuid = str(uuid4())
            await self._async_chat_log_to_cloud(
                session_uuid=session_uuid,
                version_uuid=version_details["uuid"],
                chat_log_request_list=[
                    ChatLogRequest(
                        uuid=log_uuid,
                        message=llm_response.api_response.choices[
                            0
                        ].message.model_dump(),
                        metadata=metadata,
                        api_response=api_response,
                    )
                ],
            )

            if error_occurs:
                # delete all promptmodel data in llm_response
                llm_response.raw_output = None
                llm_response.parsed_outputs = None
                llm_response.function_call = None

            llm_response.pm_detail = PMDetail(
                model=version_details["model"],
                name=self._name,
                version_uuid=str(version_details["uuid"]),
                version=version_details["version"],
                log_uuid=log_uuid,
            )
            return llm_response

        return async_wrapper

    def _wrap_chat_gen(self, gen: Callable[..., Any]) -> Callable[..., Any]:
        def wrapper(session_uuid: str, **kwargs):
            instruction, version_details, message_logs = run_async_in_sync(
                LLMProxy.fetch_chat_model(self._name, session_uuid, self.version)
            )

            call_args = self._prepare_call_args_for_chat(
                message_logs, version_details, kwargs
            )
            # Call the generator with the arguments
            stream_response: Generator[LLMStreamResponse, None, None] = gen(**call_args)

            api_response = None
            error_occurs = False
            error_log = None
            log_uuid = str(uuid4())
            for item in stream_response:
                if (
                    item.api_response and "delta" not in item.api_response.choices[0]
                ):  # only get the last api_response, not delta response
                    api_response = item.api_response

                if item.error and not error_occurs:
                    error_occurs = True
                    error_log = item.error_log

                if error_occurs:
                    # delete all promptmodel data in item
                    item.raw_output = None
                    item.parsed_outputs = None
                    item.function_call = None
                item.pm_detail = PMDetail(
                    model=version_details["model"],
                    name=self._name,
                    version_uuid=str(version_details["uuid"]),
                    version=version_details["version"],
                    log_uuid=log_uuid,
                )
                yield item

            metadata = {
                "error": error_occurs,
                "error_log": error_log,
            }
            run_async_in_sync(
                self._async_chat_log_to_cloud(
                    session_uuid=session_uuid,
                    version_uuid=version_details["uuid"],
                    chat_log_request_list=[
                        ChatLogRequest(
                            uuid=log_uuid,
                            message=api_response.choices[0].message.model_dump(),
                            metadata=metadata,
                            api_response=api_response,
                        )
                    ],
                )
            )

        return wrapper

    def _wrap_async_chat_gen(self, async_gen: Callable[..., Any]) -> Callable[..., Any]:
        async def wrapper(session_uuid: str, **kwargs):
            (
                instruction,
                version_details,
                message_logs,
            ) = await LLMProxy.fetch_chat_model(self._name, session_uuid, self.version)

            call_args = self._prepare_call_args_for_chat(
                message_logs, version_details, kwargs
            )
            # Call the generator with the arguments
            stream_response: AsyncGenerator[LLMStreamResponse, None] = async_gen(
                **call_args
            )

            api_response = None
            error_occurs = False
            error_log = None
            log_uuid = str(uuid4())
            async for item in stream_response:
                if (
                    item.api_response and "delta" not in item.api_response.choices[0]
                ):  # only get the last api_response, not delta response
                    api_response = item.api_response

                if item.error and not error_occurs:
                    error_occurs = True
                    error_log = item.error_log

                if error_occurs:
                    # delete all promptmodel data in item
                    item.raw_output = None
                    item.parsed_outputs = None
                    item.function_call = None

                item.pm_detail = PMDetail(
                    model=version_details["model"],
                    name=self._name,
                    version_uuid=str(version_details["uuid"]),
                    version=version_details["version"],
                    log_uuid=log_uuid,
                )
                yield item

            metadata = {
                "error": error_occurs,
                "error_log": error_log,
            }
            await self._async_chat_log_to_cloud(
                session_uuid=session_uuid,
                version_uuid=version_details["uuid"],
                chat_log_request_list=[
                    ChatLogRequest(
                        uuid=log_uuid,
                        message=api_response.choices[0].message.model_dump(),
                        metadata=metadata,
                        api_response=api_response,
                    )
                ],
            )

        return wrapper

    def _prepare_call_args(
        self,
        prompts: List[Dict[str, str]],
        version_detail: Dict[str, Any],
        inputs: Dict[str, Any],
        kwargs,
    ):
        stringified_inputs = {key: str(value) for key, value in inputs.items()}
        messages = [
            {
                "content": prompt["content"].format(**stringified_inputs),
                "role": prompt["role"],
            }
            for prompt in prompts
        ]
        call_args = {
            "messages": messages,
            "model": version_detail["model"] if version_detail else None,
            "parsing_type": version_detail["parsing_type"] if version_detail else None,
            "output_keys": version_detail["output_keys"] if version_detail else None,
        }
        if call_args["parsing_type"] is None:
            del call_args["parsing_type"]
            del call_args["output_keys"]

        if "functions" in kwargs:
            call_args["functions"] = kwargs["functions"]

        if "tools" in kwargs:
            call_args["tools"] = kwargs["tools"]

        if "api_key" in kwargs:
            call_args["api_key"] = kwargs["api_key"]
        return call_args

    def _prepare_call_args_for_chat(
        self,
        messages: List[Dict[str, Any]],
        version_detail: Dict[str, Any],
        kwargs,
    ):
        call_args = {}
        token_per_tools = 0
        if "functions" in kwargs:
            call_args["functions"] = kwargs["functions"]
            token_per_tools = num_tokens_from_functions_input(
                functions=kwargs["functions"],
                model=version_detail["model"] if version_detail else "gpt-3.5-turbo",
            )

        if "tools" in kwargs:
            call_args["tools"] = kwargs["tools"]
            token_per_tools = num_tokens_from_functions_input(
                functions=kwargs["tools"],
                model=version_detail["model"] if version_detail else "gpt-3.5-turbo",
            )

        # truncate messages to make length <= model's max length
        model_max_tokens = get_max_tokens(
            model=version_detail["model"] if version_detail else "gpt-3.5-turbo"
        )
        token_per_messages = num_tokens_for_messages_for_each(
            messages, version_detail["model"]
        )
        token_limit_exceeded = (
            sum(token_per_messages) + token_per_tools
        ) - model_max_tokens
        if token_limit_exceeded > 0:
            while token_limit_exceeded > 0:
                # erase the second oldest message (first one is system prompt, so it should not be erased)
                if len(messages) == 1:
                    # if there is only one message, Error cannot be solved. Just call LLM and get error response
                    break
                token_limit_exceeded -= token_per_messages[1]
                del messages[1]
                del token_per_messages[1]

        call_args["messages"] = messages
        call_args["model"] = version_detail["model"] if version_detail else None

        if "api_key" in kwargs:
            call_args["api_key"] = kwargs["api_key"]

        if "tools" in kwargs:
            call_args["tools"] = kwargs["tools"]

        return call_args

    async def _async_log_to_cloud(
        self,
        version_uuid: str,
        log_uuid: str,
        inputs: Optional[Dict] = None,
        api_response: Optional[ModelResponse] = None,
        parsed_outputs: Optional[Dict] = None,
        metadata: Optional[Dict] = None,
    ):
        config = read_config()
        if (
            "project" in config
            and "mask_inputs" in config["project"]
            and config["project"]["mask_inputs"] == True
        ):
            inputs = {key: "PRIVATE LOGGING" for key, value in inputs.items()}

        # Perform the logging asynchronously
        if api_response:
            api_response_dict = api_response.model_dump()
            api_response_dict["response_ms"] = api_response._response_ms
        else:
            api_response_dict = None
        run_log_request_body = {
            "uuid": log_uuid,
            "api_response": api_response_dict,
            "inputs": inputs,
            "parsed_outputs": parsed_outputs,
            "metadata": metadata,
        }
        res = await AsyncAPIClient.execute(
            method="POST",
            path="/run_log",
            params={
                "version_uuid": version_uuid,
            },
            json=run_log_request_body,
            use_cli_key=False,
        )
        if res.status_code != 200:
            print(f"[red]Failed to log to cloud: {res.json()}[/red]")
        return res

    async def _async_chat_log_to_cloud(
        self,
        session_uuid: str,
        version_uuid: Optional[str] = None,
        chat_log_request_list: List[ChatLogRequest] = [],
    ):
        # Perform the logging asynchronously

        res = await AsyncAPIClient.execute(
            method="POST",
            path="/chat_log",
            params={
                "session_uuid": session_uuid,
                "version_uuid": version_uuid,
            },
            json=[r.model_dump() for r in chat_log_request_list],
            use_cli_key=False,
        )
        if res.status_code != 200:
            print(f"[red]Failed to log to cloud: {res.json()}[/red]")
        return res

    async def _async_make_session_cloud(
        self,
        session_uuid: str,
        version_uuid: Optional[str] = None,
    ):
        # Perform the logging asynchronously
        res = await AsyncAPIClient.execute(
            method="POST",
            path="/make_session",
            params={
                "session_uuid": session_uuid,
                "version_uuid": version_uuid,
            },
            use_cli_key=False,
        )
        if res.status_code != 200:
            print(f"[red]Failed to make ChatSession in cloud: {res.json()}[/red]")
        return res

    def make_kwargs(self, **kwargs):
        res = {}
        for key, value in kwargs.items():
            if value is not None:
                res[key] = value
        return res

    def run(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> LLMResponse:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_method(super().run)(inputs, **kwargs)

    def arun(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> LLMResponse:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_async_method(super().arun)(inputs, **kwargs)

    def stream(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> Generator[LLMStreamResponse, None, None]:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_gen(super().stream)(inputs, **kwargs)

    def astream(
        self,
        inputs: Optional[Dict[str, Any]] = {},
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_async_gen(super().astream)(inputs, **kwargs)

    def run_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> LLMResponse:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_method(super().run_and_parse)(inputs, **kwargs)

    def arun_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> LLMResponse:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_async_method(super().arun_and_parse)(inputs, **kwargs)

    def stream_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> Generator[LLMStreamResponse, None, None]:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_gen(super().stream_and_parse)(inputs, **kwargs)

    def astream_and_parse(
        self,
        inputs: Dict[str, Any] = {},
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_async_gen(super().astream_and_parse)(inputs, **kwargs)

    def chat_run(
        self,
        session_uuid: str,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> LLMResponse:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_chat(super().run)(session_uuid, **kwargs)

    def chat_arun(
        self,
        session_uuid: str,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> LLMResponse:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_async_chat(super().arun)(session_uuid, **kwargs)

    def chat_stream(
        self,
        session_uuid: str,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> LLMResponse:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_chat_gen(super().stream)(session_uuid, **kwargs)

    def chat_astream(
        self,
        session_uuid: str,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        api_key: Optional[str] = None,
    ) -> LLMResponse:
        kwargs = self.make_kwargs(functions=functions, api_key=api_key, tools=tools)
        return self._wrap_async_chat_gen(super().astream)(session_uuid, **kwargs)

    @staticmethod
    async def fetch_prompts(
        name,
        version: Optional[Union[str, int]] = "deploy",
    ) -> Tuple[List[Dict[str, str]], Dict[str, Any]]:
        """fetch prompts.

        Args:
            name (str): name of FunctionModel

        Returns:
            Tuple[List[Dict[str, str]], Optional[Dict[str, Any]]]: (prompts, version_detail)
        """
        # Check connection activate
        config = read_config()
        if (
            "connection" in config
            and "initializing" in config["connection"]
            and config["connection"]["initializing"] == True
        ):
            return [], {}
        elif (
            "connection" in config
            and "reloading" in config["connection"]
            and config["connection"]["reloading"] == True
        ):
            return [], {}
        else:
            if (
                "project" in config
                and "use_cache" in config["project"]
                and config["project"]["use_cache"] == True
                and version == "deploy"
            ):
                cache_manager = CacheManager()
                # call update_local API in background task
                cache_update_thread = Thread(
                    target=cache_manager.cache_update_background_task, args=(config,)
                )
                cache_update_thread.daemon = True
                cache_update_thread.start()

                # get prompt from local DB by ratio
                prompt_rows, version_detail = get_deployed_prompts(name)
                if prompt_rows is None:
                    return [], {}

                return [
                    {"role": prompt.role, "content": prompt.content}
                    for prompt in prompt_rows
                ], version_detail

            else:
                try:
                    config_list = await AsyncAPIClient.execute(
                        method="GET",
                        path="/function_model_versions",
                        params={"function_model_name": name, "version": version},
                        use_cli_key=False,
                    )
                    config_list = config_list.json()
                except Exception as e:
                    raise e

                function_model_versions = [
                    x["function_model_version"] for x in config_list
                ]

                if version == "deploy":
                    for version in function_model_versions:
                        if version["is_published"] is True:
                            version["ratio"] = 1.0
                    selected_version = select_version_by_ratio(function_model_versions)
                else:
                    selected_version = function_model_versions[0]

                # config.prompts where config.function_model_version.uuid = selected_version.uuid
                prompt_rows = [
                    config["prompts"]
                    for config in config_list
                    if config["function_model_version"]["uuid"]
                    == selected_version["uuid"]
                ][0]

                # sort prompt_rows by step
                prompt_rows = sorted(prompt_rows, key=lambda prompt: prompt["step"])

                version_detail = {
                    "model": selected_version["model"],
                    "version": selected_version["version"],
                    "uuid": selected_version["uuid"],
                    "parsing_type": selected_version["parsing_type"],
                    "output_keys": selected_version["output_keys"],
                }

                if prompt_rows is None:
                    return [], {}

                return [
                    {"role": prompt["role"], "content": prompt["content"]}
                    for prompt in prompt_rows
                ], version_detail

    @staticmethod
    async def fetch_chat_model(
        name: str,
        session_uuid: Optional[str] = None,
        version: Optional[Union[str, int]] = "deploy",
    ) -> Tuple[str, Dict[str, Any], List[Dict]]:
        """fetch instruction and version detail

        Args:
            name (str): name of ChatModel

        Returns:
            Tuple[List[Dict[str, str]], Optional[Dict[str, Any]]]: (prompts, version_detail)
        """
        # Check connection activate
        config = read_config()
        if (
            "connection" in config
            and "initializing" in config["connection"]
            and config["connection"]["initializing"] == True
        ):
            return "", {}, []
        elif (
            "connection" in config
            and "reloading" in config["connection"]
            and config["connection"]["reloading"] == True
        ):
            return "", {}, []
        else:
            try:
                res_data = await AsyncAPIClient.execute(
                    method="GET",
                    path="/chat_model_versions_with_logs",
                    params={
                        "chat_model_name": name,
                        "session_uuid": session_uuid,
                        "version": version,
                    },
                    use_cli_key=False,
                )
                res_data = res_data.json()
            except Exception as e:
                raise e
            chat_model_versions = res_data["chat_model_versions"]

            if (
                session_uuid is None
            ):  # if this is the initial call for deployed chat model
                if version == "deploy":
                    for version in chat_model_versions:
                        if version["is_published"] is True:
                            version["ratio"] = 1.0
                    selected_version = select_version_by_ratio(chat_model_versions)
                else:
                    selected_version = chat_model_versions[0]
            else:
                selected_version = chat_model_versions[0]

            instruction: str = selected_version["system_prompt"]

            version_detail = {
                "model": selected_version["model"],
                "uuid": selected_version["uuid"],
                "version": selected_version["version"],
            }
            if session_uuid:
                chat_logs: List[Dict] = res_data["chat_logs"]
                chat_logs = [{"role": "system", "content": instruction}] + chat_logs
            else:
                chat_logs = []

            # delete columns which value is None in each chat log
            for chat_log in chat_logs:
                for key in list(chat_log.keys()):
                    if chat_log[key] is None:
                        del chat_log[key]

            return instruction, version_detail, chat_logs

    # @staticmethod
    # async def fetch_chat_log(
    #     session_uuid: str,
    #     version: Optional[Union[str, int]] = "deploy",
    # ) -> List[Dict[str, Any]]:
    #     """fetch conversation log for session_uuid and version detail

    #     Args:
    #         session_uuid (str): session_uuid

    #     Returns:
    #         List[Dict[str, Any]] : list of conversation log
    #     """
    #     config = read_config()
    #     if "connection" in config and config["connection"]["initializing"] == True:
    #         return []
    #     elif "connection" in config and config["connection"]["reloading"] == True:
    #         return []
    #     else:
    #         try:
    #             res_data = await AsyncAPIClient.execute(
    #                 method="GET",
    #                 path="/fetch_chat_logs",
    #                 params={"session_uuid": session_uuid},
    #                 use_cli_key=False,
    #             )
    #             res_data = res_data.json()
    #         except Exception as e:
    #             raise e

    #         # filter out unnecessary data
    #         res_data = [
    #             {
    #                 "role": message["role"],
    #                 "content": message["content"],
    #                 "function_call": message["function_call"],
    #             }
    #             for message in res_data["chat_logs"]
    #         ]
    #         return res_data



import pytest
import re
from datetime import datetime
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any
from litellm import ModelResponse

from promptmodel.llms.llm import LLM
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.types.response import (
    LLMResponse,
    LLMStreamResponse,
    Choices,
    StreamingChoices,
    Delta,
    Message,
    FunctionCall,
)
from promptmodel.types.enums import ParsingPattern, ParsingType, get_pattern_by_type


def generator_format(response: str):
    model_response = ModelResponse(stream=True)
    model_response.choices[0] = StreamingChoices(
        **{
            "delta": Delta(**{"content": response}),
            "finish_reason": None,
        }
    )
    return model_response


def string_to_generator(input_string: str):
    def generator():
        for char in input_string:
            yield generator_format(char)

    return generator()


import random


def random_chunk_generator(input_string: str):
    def generator():
        start = 0
        while start < len(input_string):
            chunk_size = random.randint(1, 3)
            end = min(start + chunk_size, len(input_string))
            yield generator_format(input_string[start:end])
            start = end

    return generator()


async def string_to_agenerator(input_string: str):
    async def agenerator():
        for char in input_string:
            yield generator_format(char)

    return agenerator()


import random


async def random_chunk_agenerator(input_string: str):
    async def agenerator():
        start = 0
        while start < len(input_string):
            chunk_size = random.randint(1, 3)
            end = min(start + chunk_size, len(input_string))
            yield generator_format(input_string[start:end])
            start = end

    return agenerator()


class ResultClass:
    raw_output: str
    parsed_outputs: Dict[str, str]
    parse_success: bool

    def __init__(self, raw_output, parsed_outputs, parse_success):
        self.raw_output = raw_output
        self.parsed_outputs = parsed_outputs
        self.parse_success = parse_success


bracket_allow_example_gen = string_to_generator("[key type=str]ab[/key]")
bracket_allow_result = ResultClass("[key type=str]ab[/key]", {"key": "ab"}, True)

bracket_forbidden_example_gen = string_to_generator("[key type=str]ab[/not key]")
bracket_forbidden_result = ResultClass("[key type=str]ab[/not key]", {}, False)
bracket_forbidden_stream_result = ResultClass(
    "[key type=str]ab[/not key]", {"key": "ab[/not key]"}, False
)

double_allow_example_gen = string_to_generator(
    "[[key type=str]]abc[not key]xy[/not key]de[[/key]]"
)
double_allow_result = ResultClass(
    "[[key type=str]]abc[not key]xy[/not key]de[[/key]]",
    {"key": "abc[not key]xy[/not key]de"},
    True,
)

double_forbidden_example_gen = string_to_generator(
    "[[key type=str]]abc[not key]xy[/not key]de[[/not key]]"
)
double_forbidden_result = ResultClass(
    "[[key type=str]]abc[not key]xy[/not key]de[[/not key]]", {}, False
)
double_forbidden_stream_result = ResultClass(
    "[[key type=str]]abc[not key]xy[/not key]de[[/not key]]",
    {"key": "abc[not key]xy[/not key]de"},
    False,
)


def generator_processor(generator: Generator):
    raw_output = ""
    parsed_outputs = {}
    parse_success = True
    for chunk in generator:
        # print(chunk)
        if type(chunk) == str:
            raw_output += chunk
        elif type(chunk) == dict:
            for key, value in chunk.items():
                if key in parsed_outputs:
                    parsed_outputs[key] += value
                else:
                    parsed_outputs[key] = value
        elif type(chunk) == bool:
            parse_success = chunk
        elif type(chunk) == LLMResponse:
            if chunk.raw_output:
                raw_output += chunk.raw_output
            if chunk.parsed_outputs:
                for key, value in chunk.parsed_outputs.items():
                    if key in parsed_outputs:
                        parsed_outputs[key] += value
                    else:
                        parsed_outputs[key] = value
            if parse_success is True and chunk.error is True:
                parse_success = False
        elif type(chunk) == LLMStreamResponse:
            if chunk.raw_output:
                raw_output += chunk.raw_output
            if chunk.parsed_outputs:
                for key, value in chunk.parsed_outputs.items():
                    if key in parsed_outputs:
                        parsed_outputs[key] += value
                    else:
                        parsed_outputs[key] = value
            if parse_success is True and chunk.error is True:
                parse_success = False

    return raw_output, parsed_outputs, parse_success


async def agenerator_processor(generator: AsyncGenerator):
    raw_output = ""
    parsed_outputs = {}
    parse_success = True
    async for chunk in generator:
        # print(chunk)
        if type(chunk) == str:
            raw_output += chunk
        elif type(chunk) == dict:
            for key, value in chunk.items():
                if key in parsed_outputs:
                    parsed_outputs[key] += value
                else:
                    parsed_outputs[key] = value
        elif type(chunk) == bool:
            parse_success = chunk
        elif type(chunk) == LLMResponse:
            if chunk.raw_output:
                raw_output += chunk.raw_output
            if chunk.parsed_outputs:
                for key, value in chunk.parsed_outputs.items():
                    if key in parsed_outputs:
                        parsed_outputs[key] += value
                    else:
                        parsed_outputs[key] = value
            if parse_success is True and chunk.error is True:
                parse_success = False
        elif type(chunk) == LLMStreamResponse:
            if chunk.raw_output:
                raw_output += chunk.raw_output
            if chunk.parsed_outputs:
                for key, value in chunk.parsed_outputs.items():
                    if key in parsed_outputs:
                        parsed_outputs[key] += value
                    else:
                        parsed_outputs[key] = value
            if parse_success is True and chunk.error is True:
                parse_success = False

    return raw_output, parsed_outputs, parse_success


def test_sp_generator(mocker):
    llm = LLM()

    raw_output, parsed_outputs, parse_success = generator_processor(
        llm.__single_type_sp_generator__(
            None,
            bracket_allow_example_gen,
            ParsingType.SQUARE_BRACKET.value,
            datetime.now(),
        )
    )
    print(raw_output, parsed_outputs, parse_success)

    assert (
        raw_output == bracket_allow_result.raw_output
    ), f"raw output error : {raw_output}"
    assert (
        parsed_outputs == bracket_allow_result.parsed_outputs
    ), f"parsed output error : {parsed_outputs}"
    assert (
        parse_success == bracket_allow_result.parse_success
    ), f"parse success mismatch : {parse_success}"

    raw_output, parsed_outputs, parse_success = generator_processor(
        llm.__single_type_sp_generator__(
            None,
            bracket_forbidden_example_gen,
            ParsingType.SQUARE_BRACKET.value,
            datetime.now(),
        )
    )
    print(raw_output, parsed_outputs, parse_success)
    assert (
        raw_output == bracket_forbidden_stream_result.raw_output
    ), f"raw output error : {raw_output}"
    assert (
        parsed_outputs == bracket_forbidden_stream_result.parsed_outputs
    ), f"parsed output error : {parsed_outputs}"
    assert (
        parse_success == bracket_forbidden_stream_result.parse_success
    ), f"parse success mismatch : {parse_success}"
    # test Random Generator 5 times
    for i in range(5):
        raw_output, parsed_outputs, parse_success = generator_processor(
            llm.__single_type_sp_generator__(
                None,
                random_chunk_generator("[key type=str]ab[/key]"),
                ParsingType.SQUARE_BRACKET.value,
                datetime.now(),
            )
        )
        print(raw_output, parsed_outputs, parse_success)

        assert (
            raw_output == bracket_allow_result.raw_output
        ), f"raw output error : {raw_output}"
        assert (
            parsed_outputs == bracket_allow_result.parsed_outputs
        ), f"parsed output error : {parsed_outputs}"
        assert (
            parse_success == bracket_allow_result.parse_success
        ), f"parse success mismatch : {parse_success}"

        raw_output, parsed_outputs, parse_success = generator_processor(
            llm.__single_type_sp_generator__(
                None,
                random_chunk_generator("[key type=str]ab[/not key]"),
                ParsingType.SQUARE_BRACKET.value,
                datetime.now(),
            )
        )
        print(raw_output, parsed_outputs, parse_success)
        assert (
            raw_output == bracket_forbidden_stream_result.raw_output
        ), f"raw output error : {raw_output}"
        assert (
            parsed_outputs == bracket_forbidden_stream_result.parsed_outputs
        ), f"parsed output error : {parsed_outputs}"
        assert (
            parse_success == bracket_forbidden_stream_result.parse_success
        ), f"parse success mismatch : {parse_success}"


def test_dp_generator(mocker):
    llm = LLM()
    raw_output, parsed_outputs, parse_success = generator_processor(
        llm.__double_type_sp_generator__(
            None,
            double_allow_example_gen,
            ParsingType.DOUBLE_SQUARE_BRACKET.value,
            datetime.now(),
        )
    )
    print(raw_output, parsed_outputs, parse_success)

    assert (
        raw_output == double_allow_result.raw_output
    ), f"raw output error : {raw_output}"
    assert (
        parsed_outputs == double_allow_result.parsed_outputs
    ), f"parsed output error : {parsed_outputs}"
    assert (
        parse_success == double_allow_result.parse_success
    ), f"parse success mismatch : {parse_success}"

    raw_output, parsed_outputs, parse_success = generator_processor(
        llm.__double_type_sp_generator__(
            None,
            double_forbidden_example_gen,
            ParsingType.DOUBLE_SQUARE_BRACKET.value,
            datetime.now(),
        )
    )
    print(raw_output, parsed_outputs, parse_success)
    assert (
        raw_output == double_forbidden_stream_result.raw_output
    ), f"raw output error : {raw_output}"
    assert (
        parsed_outputs == double_forbidden_stream_result.parsed_outputs
    ), f"parsed output error : {parsed_outputs}"
    assert (
        parse_success == double_forbidden_stream_result.parse_success
    ), f"parse success mismatch : {parse_success}"

    # test random generator 5 times
    for i in range(5):
        raw_output, parsed_outputs, parse_success = generator_processor(
            llm.__double_type_sp_generator__(
                None,
                random_chunk_generator(
                    "[[key type=str]]abc[not key]xy[/not key]de[[/key]]"
                ),
                ParsingType.DOUBLE_SQUARE_BRACKET.value,
                datetime.now(),
            )
        )
        print(raw_output, parsed_outputs, parse_success)

        assert (
            raw_output == double_allow_result.raw_output
        ), f"raw output error : {raw_output}"
        assert (
            parsed_outputs == double_allow_result.parsed_outputs
        ), f"parsed output error : {parsed_outputs}"
        assert (
            parse_success == double_allow_result.parse_success
        ), f"parse success mismatch : {parse_success}"

        raw_output, parsed_outputs, parse_success = generator_processor(
            llm.__double_type_sp_generator__(
                None,
                random_chunk_generator(
                    "[[key type=str]]abc[not key]xy[/not key]de[[/not key]]"
                ),
                ParsingType.DOUBLE_SQUARE_BRACKET.value,
                datetime.now(),
            )
        )
        print(raw_output, parsed_outputs, parse_success)
        assert (
            raw_output == double_forbidden_stream_result.raw_output
        ), f"raw output error : {raw_output}"
        assert (
            parsed_outputs == double_forbidden_stream_result.parsed_outputs
        ), f"parsed output error : {parsed_outputs}"
        assert (
            parse_success == double_forbidden_stream_result.parse_success
        ), f"parse success mismatch : {parse_success}"


@pytest.mark.asyncio
async def test_sp_agenerator(mocker):
    llm = LLM()
    bracket_allow_example_agen = await string_to_agenerator("[key type=str]ab[/key]")
    bracket_forbidden_example_agen = await string_to_agenerator(
        "[key type=str]ab[/not key]"
    )

    raw_output, parsed_outputs, parse_success = await agenerator_processor(
        llm.__single_type_sp_agenerator__(
            None,
            bracket_allow_example_agen,
            ParsingType.SQUARE_BRACKET.value,
            datetime.now(),
        )
    )
    print(raw_output, parsed_outputs, parse_success)

    assert (
        raw_output == bracket_allow_result.raw_output
    ), f"raw output error : {raw_output}"
    assert (
        parsed_outputs == bracket_allow_result.parsed_outputs
    ), f"parsed output error : {parsed_outputs}"
    assert (
        parse_success == bracket_allow_result.parse_success
    ), f"parse success mismatch : {parse_success}"

    raw_output, parsed_outputs, parse_success = await agenerator_processor(
        llm.__single_type_sp_agenerator__(
            None,
            bracket_forbidden_example_agen,
            ParsingType.SQUARE_BRACKET.value,
            datetime.now(),
        )
    )
    print(raw_output, parsed_outputs, parse_success)
    assert (
        raw_output == bracket_forbidden_stream_result.raw_output
    ), f"raw output error : {raw_output}"
    assert (
        parsed_outputs == bracket_forbidden_stream_result.parsed_outputs
    ), f"parsed output error : {parsed_outputs}"
    assert (
        parse_success == bracket_forbidden_stream_result.parse_success
    ), f"parse success mismatch : {parse_success}"
    # test Random Generator 5 times
    for i in range(5):
        raw_output, parsed_outputs, parse_success = await agenerator_processor(
            llm.__single_type_sp_agenerator__(
                None,
                await random_chunk_agenerator("[key type=str]ab[/key]"),
                ParsingType.SQUARE_BRACKET.value,
                datetime.now(),
            )
        )
        print(raw_output, parsed_outputs, parse_success)

        assert (
            raw_output == bracket_allow_result.raw_output
        ), f"raw output error : {raw_output}"
        assert (
            parsed_outputs == bracket_allow_result.parsed_outputs
        ), f"parsed output error : {parsed_outputs}"
        assert (
            parse_success == bracket_allow_result.parse_success
        ), f"parse success mismatch : {parse_success}"

        raw_output, parsed_outputs, parse_success = await agenerator_processor(
            llm.__single_type_sp_agenerator__(
                None,
                await random_chunk_agenerator("[key type=str]ab[/not key]"),
                ParsingType.SQUARE_BRACKET.value,
                datetime.now(),
            )
        )
        print(raw_output, parsed_outputs, parse_success)
        assert (
            raw_output == bracket_forbidden_stream_result.raw_output
        ), f"raw output error : {raw_output}"
        assert (
            parsed_outputs == bracket_forbidden_stream_result.parsed_outputs
        ), f"parsed output error : {parsed_outputs}"
        assert (
            parse_success == bracket_forbidden_stream_result.parse_success
        ), f"parse success mismatch : {parse_success}"


@pytest.mark.asyncio
async def test_dp_agenerator(mocker):
    llm = LLM()
    double_allow_example_agen = await string_to_agenerator(
        "[[key type=str]]abc[not key]xy[/not key]de[[/key]]"
    )
    double_forbidden_example_agen = await string_to_agenerator(
        "[[key type=str]]abc[not key]xy[/not key]de[[/not key]]"
    )

    raw_output, parsed_outputs, parse_success = await agenerator_processor(
        llm.__double_type_sp_agenerator__(
            None,
            double_allow_example_agen,
            ParsingType.DOUBLE_SQUARE_BRACKET.value,
            datetime.now(),
        )
    )
    print(raw_output, parsed_outputs, parse_success)

    assert (
        raw_output == double_allow_result.raw_output
    ), f"raw output error : {raw_output}"
    assert (
        parsed_outputs == double_allow_result.parsed_outputs
    ), f"parsed output error : {parsed_outputs}"
    assert (
        parse_success == double_allow_result.parse_success
    ), f"parse success mismatch : {parse_success}"

    raw_output, parsed_outputs, parse_success = await agenerator_processor(
        llm.__double_type_sp_agenerator__(
            None,
            double_forbidden_example_agen,
            ParsingType.DOUBLE_SQUARE_BRACKET.value,
            datetime.now(),
        )
    )
    print(raw_output, parsed_outputs, parse_success)
    assert (
        raw_output == double_forbidden_stream_result.raw_output
    ), f"raw output error : {raw_output}"
    assert (
        parsed_outputs == double_forbidden_stream_result.parsed_outputs
    ), f"parsed output error : {parsed_outputs}"
    assert (
        parse_success == double_forbidden_stream_result.parse_success
    ), f"parse success mismatch : {parse_success}"

    # test random generator 5 times
    for i in range(5):
        raw_output, parsed_outputs, parse_success = await agenerator_processor(
            llm.__double_type_sp_agenerator__(
                None,
                await random_chunk_agenerator(
                    "[[key type=str]]abc[not key]xy[/not key]de[[/key]]"
                ),
                ParsingType.DOUBLE_SQUARE_BRACKET.value,
                datetime.now(),
            )
        )
        print(raw_output, parsed_outputs, parse_success)

        assert (
            raw_output == double_allow_result.raw_output
        ), f"raw output error : {raw_output}"
        assert (
            parsed_outputs == double_allow_result.parsed_outputs
        ), f"parsed output error : {parsed_outputs}"
        assert (
            parse_success == double_allow_result.parse_success
        ), f"parse success mismatch : {parse_success}"

        raw_output, parsed_outputs, parse_success = await agenerator_processor(
            llm.__double_type_sp_agenerator__(
                None,
                await random_chunk_agenerator(
                    "[[key type=str]]abc[not key]xy[/not key]de[[/not key]]"
                ),
                ParsingType.DOUBLE_SQUARE_BRACKET.value,
                datetime.now(),
            )
        )
        print(raw_output, parsed_outputs, parse_success)
        assert (
            raw_output == double_forbidden_stream_result.raw_output
        ), f"raw output error : {raw_output}"
        assert (
            parsed_outputs == double_forbidden_stream_result.parsed_outputs
        ), f"parsed output error : {parsed_outputs}"
        assert (
            parse_success == double_forbidden_stream_result.parse_success
        ), f"parse success mismatch : {parse_success}"


@pytest.mark.asyncio
async def test_run_and_parsing(mocker):
    llm = LLM()

    mock_response = ModelResponse()
    mock_response.choices = [
        Choices(
            **{
                "index": 0,
                "message": Message(
                    **{
                        "role": "assistant",
                        "content": "[key type=str]ab[/key]",
                    }
                ),
                "finish_reason": "function_call",
            }
        )
    ]
    mock_completion = mocker.patch(
        "promptmodel.llms.llm.completion", return_value=mock_response
    )

    # success case
    res: LLMResponse = llm.run_and_parse(
        messages=[{"role": "user", "content": "What is the weather like in Boston?"}],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
         
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.parsed_outputs == {
        "key": "ab"
    }, f"parsed output error : {res.parsed_outputs}"

    # failed case
    res: LLMResponse = llm.run_and_parse(
        messages=[{"role": "user", "content": "What is the weather like in Boston?"}],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
         
        output_keys=["key", "key2"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.parsed_outputs == {
        "key": "ab"
    }, f"parsed output error : {res.parsed_outputs}"
    assert res.error == True, f"error mismatch : {res.error}"

    # error case, raise in __parse_output_pattern__
    mock_completion = mocker.patch(
        "promptmodel.llms.llm.completion", return_value=mock_response
    )
    with mocker.patch(
        "promptmodel.llms.llm.LLM.__parse_output_pattern__",
        side_effect=Exception("test"),
    ):
        res: LLMResponse = llm.run_and_parse(
            messages=[
                {"role": "user", "content": "What is the weather like in Boston?"}
            ],
            parsing_type=ParsingType.SQUARE_BRACKET.value,
             
            output_keys=["key"],
            model="gpt-3.5-turbo",
        )
        mock_completion.assert_called_once()
        mock_completion.reset_mock()
        # should have api response
        assert (
            res.api_response is not None
        ), f"api response mismatch : {res.api_response}"
        assert res.error == True, f"error mismatch : {res.error}"

    # success case with function call
    mock_response.choices[0].message.function_call = FunctionCall(
        **{
            "name": "get_current_weather",
            "arguments": '{\n  "location": "Boston, MA"\n}',
        }
    )
    res: LLMResponse = llm.run_and_parse(
        messages=[{"role": "user", "content": "What is the weather like in Boston?"}],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
         
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert (
        res.parsed_outputs is None
    ), f"If function call, do not parse but : {res.parsed_outputs}"
    assert res.function_call == FunctionCall(
        **{
            "name": "get_current_weather",
            "arguments": '{\n  "location": "Boston, MA"\n}',
        }
    ), f"function call mismatch : {res.function_call}"

    # success case because function_call, did not parse
    mocker.patch(
        "promptmodel.llms.llm.LLM.__parse_output_pattern__",
        side_effect=Exception("test"),
    )
    res: LLMResponse = llm.run_and_parse(
        messages=[{"role": "user", "content": "What is the weather like in Boston?"}],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
         
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.api_response is not None, f"api response mismatch : {res.api_response}"
    assert res.error == False, f"error mismatch : {res.error}"

    # error case, raise error in completion
    mock_completion = mocker.patch(
        "promptmodel.llms.llm.completion",
        side_effect=Exception("test"),
        return_value=None,
    )
    res: LLMResponse = llm.run_and_parse(
        messages=[{"role": "user", "content": "What is the weather like in Boston?"}],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
         
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.api_response is None, f"api response mismatch : {res.api_response}"
    assert res.error == True, f"error mismatch : {res.error}"


@pytest.mark.asyncio
async def test_arun_and_parsing(mocker):
    llm = LLM()

    mock_response = ModelResponse()
    mock_response.choices = [
        Choices(
            **{
                "index": 0,
                "message": Message(
                    **{
                        "role": "assistant",
                        "content": "[key type=str]ab[/key]",
                    }
                ),
                "finish_reason": "function_call",
            }
        )
    ]
    mock_completion = mocker.patch(
        "promptmodel.llms.llm.acompletion", return_value=mock_response
    )

    # success case
    res: LLMResponse = await llm.arun_and_parse(
        messages=[{"role": "user", "content": "What is the weather like in Boston?"}],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
         
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.parsed_outputs == {
        "key": "ab"
    }, f"parsed output error : {res.parsed_outputs}"

    # failed case
    res: LLMResponse = await llm.arun_and_parse(
        messages=[{"role": "user", "content": "What is the weather like in Boston?"}],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
         
        output_keys=["key", "key2"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.parsed_outputs == {
        "key": "ab"
    }, f"parsed output error : {res.parsed_outputs}"
    assert res.error == True, f"error mismatch : {res.error}"

    # error case, raise in __parse_output_pattern__
    mock_completion = mocker.patch(
        "promptmodel.llms.llm.acompletion", return_value=mock_response
    )
    with mocker.patch(
        "promptmodel.llms.llm.LLM.__parse_output_pattern__",
        side_effect=Exception("test"),
    ):
        res: LLMResponse = await llm.arun_and_parse(
            messages=[
                {"role": "user", "content": "What is the weather like in Boston?"}
            ],
            parsing_type=ParsingType.SQUARE_BRACKET.value,
             
            output_keys=["key"],
            model="gpt-3.5-turbo",
        )
        mock_completion.assert_called_once()
        mock_completion.reset_mock()
        # should have api response
        assert (
            res.api_response is not None
        ), f"api response mismatch : {res.api_response}"
        assert res.error == True, f"error mismatch : {res.error}"

    # success case with function call
    mock_response.choices[0].message.function_call = FunctionCall(
        **{
            "name": "get_current_weather",
            "arguments": '{\n  "location": "Boston, MA"\n}',
        }
    )
    res: LLMResponse = await llm.arun_and_parse(
        messages=[{"role": "user", "content": "What is the weather like in Boston?"}],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
         
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert (
        res.parsed_outputs is None
    ), f"If function call, do not parse but : {res.parsed_outputs}"
    assert res.function_call == FunctionCall(
        **{
            "name": "get_current_weather",
            "arguments": '{\n  "location": "Boston, MA"\n}',
        }
    ), f"function call mismatch : {res.function_call}"

    # success case because function_call, did not parse
    mocker.patch(
        "promptmodel.llms.llm.LLM.__parse_output_pattern__",
        side_effect=Exception("test"),
    )
    res: LLMResponse = await llm.arun_and_parse(
        messages=[{"role": "user", "content": "What is the weather like in Boston?"}],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
         
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.api_response is not None, f"api response mismatch : {res.api_response}"
    assert res.error == False, f"error mismatch : {res.error}"

    # error case, raise error in completion
    mock_completion = mocker.patch(
        "promptmodel.llms.llm.acompletion",
        side_effect=Exception("test"),
        return_value=None,
    )
    res: LLMResponse = await llm.arun_and_parse(
        messages=[{"role": "user", "content": "What is the weather like in Boston?"}],
        parsing_type=ParsingType.SQUARE_BRACKET.value,
         
        output_keys=["key"],
        model="gpt-3.5-turbo",
    )
    mock_completion.assert_called_once()
    mock_completion.reset_mock()
    assert res.api_response is None, f"api response mismatch : {res.api_response}"
    assert res.error == True, f"error mismatch : {res.error}"



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from promptmodel.websocket.websocket_client import DevWebsocketClient
from promptmodel.dev_app import DevApp


async def echo_coroutine(*args, **kwargs):
    # print(args, kwargs)
    return args, kwargs


@pytest.fixture
def websocket_client():
    websocket_client = DevWebsocketClient(_devapp=DevApp())
    return websocket_client


@pytest.fixture
def mock_websocket():
    # 모의 WebSocketClientProtocol 객체 생성
    mock_websocket = AsyncMock()

    async def aenter(self):
        return self

    async def aexit(self, exc_type, exc_value, traceback):
        pass

    mock_websocket.__aenter__ = aenter
    mock_websocket.__aexit__ = aexit
    mock_websocket.recv = AsyncMock(return_value='{"key" : "value"}')
    mock_websocket.send = AsyncMock()

    return mock_websocket


@pytest.fixture
def mock_json_dumps():
    mock_json_dumps = MagicMock()
    # it should return exactly the same as the input
    mock_json_dumps.side_effect = lambda data, *args, **kwargs: data
    return mock_json_dumps



import datetime
import json

from uuid import uuid4
from peewee import (
    CharField,
    DateTimeField,
    IntegerField,
    ForeignKeyField,
    BooleanField,
    UUIDField,
    TextField,
    AutoField,
    FloatField,
    Check,
)


class JSONField(TextField):
    def db_value(self, value):
        return json.dumps(value)

    def python_value(self, value):
        return json.loads(value)


# class DeployedChatModel(BaseModel):
#     uuid = UUIDField(unique=True, default=uuid4)
#     name = CharField()


# class DeployedChatModelVersion(BaseModel):
#     uuid = UUIDField(unique=True, default=uuid4)
#     from_uuid = UUIDField(null=True)
#     chat_model_uuid = ForeignKeyField(
#         DeployedChatModel,
#         field=DeployedChatModel.uuid,
#         backref="versions",
#         on_delete="CASCADE",
#     )
#     model = CharField()
#     is_published = BooleanField(default=False)
#     is_ab_test = BooleanField(default=False)
#     ratio = FloatField(null=True)
#     system_prompt = JSONField(null=True, default={})
#     functions = JSONField(default=[])



"""CLI for project management."""

from promptmodel.apis.base import APIClient
import typer
from rich import print

from promptmodel.utils.config_utils import read_config, upsert_config
from ..utils import get_org

app = typer.Typer(no_args_is_help=True, short_help="Manage Client projects.")


@app.command()
def list():
    """List all projects."""
    config = read_config()
    org = get_org(config)
    projects = APIClient.execute(
        method="GET",
        path="/projects",
        params={"organization_id": org["organization_id"]},
    ).json()
    print("\nProjects:")
    for project in projects:
        print(f"📌 {project['name']} ({project['version']})")
        if project["description"]:
            print(f"   {project['description']}\n")



from importlib import resources
import typer
from rich import print
from promptmodel.constants import (
    PROMPTMODEL_DEV_FILENAME,
    PROMPTMODEL_DEV_STARTER_FILENAME,
)


def init(from_cli: bool = True):
    """Initialize a new promptmodel project."""
    import os

    if not os.path.exists(PROMPTMODEL_DEV_FILENAME):
        # Read the content from the source file
        content = resources.read_text("promptmodel", PROMPTMODEL_DEV_STARTER_FILENAME)

        # Write the content to the target file
        with open(PROMPTMODEL_DEV_FILENAME, "w") as target_file:
            target_file.write(content)
        print(
            "[violet][bold]promptmodel_dev.py[/bold][/violet] was successfully created!"
        )
        print(
            "Add promptmodels in your code, then run [violet][bold]prompt dev[/bold][/violet] to start engineering prompts."
        )
    elif from_cli:
        print(
            "[yellow]promptmodel_dev.py[/yellow] was already initialized in this directory."
        )
        print(
            "Run [violet][bold]prompt dev[/bold][/violet] to start engineering prompts."
        )


app = typer.Typer(invoke_without_command=True, callback=init)



import os
import asyncio
from typing import Any, Dict
import yaml
from functools import wraps

CONFIG_FILE = "./.promptmodel/config.yaml"


def read_config():
    """
    Reads the configuration from the given filename.

    :return: A dictionary containing the configuration.
    """
    if not os.path.exists(CONFIG_FILE):
        return {}

    with open(CONFIG_FILE, "r") as file:
        config = yaml.safe_load(file) or {}
    return config


def merge_dict(d1: Dict[str, Any], d2: Dict[str, Any]):
    """
    Merge two dictionaries recursively.

    :param d1: The first dictionary.
    :param d2: The second dictionary.
    :return: The merged dictionary.
    """
    for key, value in d2.items():
        if key in d1 and isinstance(d1[key], dict) and isinstance(value, dict):
            d1[key] = merge_dict(d1[key], value)
        else:
            d1[key] = value
    return d1


def upsert_config(new_config: Dict[str, Any], section: str = None):
    """
    Upserts the given configuration file with the given configuration.

    :param new_config: A dictionary containing the new configuration.
    :param section: The section of the configuration to update.
    """
    config = read_config()
    if section:
        config_section = config.get(section, {})
        new_config = {section: merge_dict(config_section, new_config)}
    config = merge_dict(config, new_config)
    # If . directory does not exist, create it
    if not os.path.exists("./.promptmodel"):
        os.mkdir("./.promptmodel")

    with open(CONFIG_FILE, "w") as file:
        yaml.safe_dump(config, file, default_flow_style=False)


def check_connection_status_decorator(method):
    if asyncio.iscoroutinefunction(method):

        @wraps(method)
        async def async_wrapper(self, *args, **kwargs):
            config = read_config()
            if "connection" in config and (
                (
                    "initializing" in config["connection"]
                    and config["connection"]["initializing"]
                )
                or (
                    "reloading" in config["connection"]
                    and config["connection"]["reloading"]
                )
            ):
                return
            else:
                if "config" not in kwargs:
                    kwargs["config"] = config
                return await method(self, *args, **kwargs)

        # async_wrapper.__name__ = method.__name__
        # async_wrapper.__doc__ = method.__doc__
        return async_wrapper
    else:

        @wraps(method)
        def wrapper(self, *args, **kwargs):
            config = read_config()
            if "connection" in config and (
                (
                    "initializing" in config["connection"]
                    and config["connection"]["initializing"]
                )
                or (
                    "reloading" in config["connection"]
                    and config["connection"]["reloading"]
                )
            ):
                return
            else:
                return method(self, *args, **kwargs)

        # wrapper.__name__ = method.__name__
        # wrapper.__doc__ = method.__doc__
        return wrapper



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

import asyncio
from typing import Optional
from uuid import uuid4
from dataclasses import dataclass
from websockets.exceptions import ConnectionClosedOK
from promptmodel.websocket.websocket_client import DevWebsocketClient
from promptmodel.types.enums import LocalTask, ParsingType, LocalTaskErrorType
from promptmodel.types.response import FunctionSchema


@dataclass
class FunctionModelInterface:
    name: str
    default_model: str = "gpt-3.5-turbo"


def get_current_weather(location: str, unit: Optional[str] = "celsius"):
    return "13 degrees celsius"


get_current_weather_desc = {
    "name": "get_current_weather",
    "description": "Get the current weather in a given location",
    "parameters": {
        "type": "object",
        "properties": {
            "location": {
                "type": "string",
                "description": "The city and state, e.g. San Francisco, CA",
            },
            "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
        },
        "required": ["location"],
    },
}


@pytest.mark.asyncio
async def test_run_model_function_call(
    mocker,
    websocket_client: DevWebsocketClient,
    mock_websocket: AsyncMock,
    mock_json_dumps: MagicMock,
):
    websocket_client._devapp.function_models = [FunctionModelInterface("test_module")]

    websocket_client._devapp.functions = {
        "get_current_weather": {
            "schema": FunctionSchema(**get_current_weather_desc),
            "function": get_current_weather,
        }
    }

    function_schemas_in_db = [FunctionSchema(**get_current_weather_desc).model_dump()]

    function_schemas_in_db[0]["mock_response"] = "13"

    mocker.patch("promptmodel.websocket.websocket_client.json.dumps", mock_json_dumps)

    # success case
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_PROMPT_MODEL,
            "messages_for_run": [
                {
                    "role": "system",
                    "content": "You are a helpful assistant.",
                    "step": 1,
                },
                {
                    "role": "user",
                    "content": "What is the weather like in Boston?",
                    "step": 2,
                },
            ],
            "model": "gpt-3.5-turbo",
            "parsing_type": None,
            "output_keys": None,
            "functions": ["get_current_weather"],
            "function_schemas": function_schemas_in_db,
        },
        ws=mock_websocket,
    )
    call_args_list = mock_websocket.send.call_args_list
    # print(call_args_list)
    data = [arg.args[0] for arg in call_args_list]

    assert len([d for d in data if d["status"] == "failed"]) == 0
    assert len([d for d in data if d["status"] == "completed"]) == 1
    assert len([d for d in data if "function_response" in d]) == 1
    assert [d for d in data if "function_response" in d][0]["function_response"][
        "name"
    ] == "get_current_weather"
    assert len([d for d in data if "function_call" in d]) == 1
    assert len([d for d in data if "raw_output" in d]) == 0
    mock_websocket.send.reset_mock()
    function_schemas_in_db[0]["mock_response"] = "13"

    print(
        "======================================================================================="
    )

    # success case with no function call
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_PROMPT_MODEL,
            "messages_for_run": [
                {
                    "role": "system",
                    "content": "You are a helpful assistant.",
                    "step": 1,
                },
                {
                    "role": "user",
                    "content": "What is the weather like in Boston?",
                    "step": 2,
                },
            ],
            "model": "gpt-3.5-turbo",
            "parsing_type": None,
            "output_keys": None,
        },
        ws=mock_websocket,
    )
    call_args_list = mock_websocket.send.call_args_list
    # print(call_args_list)
    data = [arg.args[0] for arg in call_args_list]

    assert len([d for d in data if "function_call" in d]) == 0
    assert len([d for d in data if "raw_output" in d]) > 0
    assert len([d for d in data if d["status"] == "failed"]) == 0
    assert len([d for d in data if d["status"] == "completed"]) == 1
    mock_websocket.send.reset_mock()

    print(
        "======================================================================================="
    )

    #     system_prompt_with_format = """
    # You are a helpful assistant.

    # This is your output format. Keep the string between "[" and "]" as same as given below. You should response content first before call functions. You MUST FOLLOW this output format.
    # Output Format:
    # [response type=str]
    # (value here)
    # [/response]
    #     """
    #     await websocket_client._DevWebsocketClient__handle_message(
    #         message={
    #             "type": LocalTask.RUN_PROMPT_MODEL,
    #             "messages_for_run": [
    #                 {"role": "system", "content": system_prompt_with_format, "step": 1},
    #                 {
    #                     "role": "user",
    #                     "content": "What is the weather like in Boston?",
    #                     "step": 2,
    #                 },
    #             ],
    #             "model": "gpt-4-1106-preview",
    #             "uuid": None,
    #             "from_uuid": None,
    #             "parsing_type": ParsingType.SQUARE_BRACKET.value,
    #             "output_keys": ["response"],
    #             "functions": ["get_current_weather"],
    #             "function_schemas": function_schemas_in_db,
    #         },
    #         ws=mock_websocket,
    #     )
    #     call_args_list = mock_websocket.send.call_args_list
    #     print(call_args_list)
    #     data = [arg.args[0] for arg in call_args_list]

    #     assert len([d for d in data if d["status"] == "failed"]) == 0
    #     assert len([d for d in data if d["status"] == "completed"]) == 1
    #     assert len([d for d in data if "function_response" in d]) == 1
    #     assert [d for d in data if "function_response" in d][0]["function_response"][
    #         "name"
    #     ] == "get_current_weather"
    #     assert len([d for d in data if "function_call" in d]) == 1
    #     assert len([d for d in data if "raw_output" in d]) > 0
    #     assert len([d for d in data if "parsed_outputs" in d]) > 0
    #     assert (
    #         len(
    #             list(
    #                 set([d["parsed_outputs"].keys() for d in data if "parsed_outputs" in d])
    #             )
    #         )
    #         == 1
    #     )
    #     mock_websocket.send.reset_mock()
    #     function_schemas_in_db[0]["mock_response"] = "13"

    print(
        "======================================================================================="
    )

    # FUNCTION_CALL_FAILED_ERROR case
    def error_raise_function(*args, **kwargs):
        raise Exception("error")

    websocket_client._devapp.functions = {
        "get_current_weather": {
            "schema": FunctionSchema(**get_current_weather_desc),
            "function": error_raise_function,
        }
    }

    # success case
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_PROMPT_MODEL,
            "messages_for_run": [
                {
                    "role": "system",
                    "content": "You are a helpful assistant.",
                    "step": 1,
                },
                {
                    "role": "user",
                    "content": "What is the weather like in Boston?",
                    "step": 2,
                },
            ],
            "model": "gpt-3.5-turbo",
            "parsing_type": None,
            "output_keys": None,
            "functions": ["get_current_weather"],
            "function_schemas": function_schemas_in_db,
        },
        ws=mock_websocket,
    )
    call_args_list = mock_websocket.send.call_args_list
    # print(call_args_list)
    data = [arg.args[0] for arg in call_args_list]

    assert len([d for d in data if d["status"] == "failed"]) == 1
    assert [d for d in data if d["status"] == "failed"][0][
        "error_type"
    ] == LocalTaskErrorType.FUNCTION_CALL_FAILED_ERROR.value
    assert len([d for d in data if d["status"] == "completed"]) == 0
    assert len([d for d in data if "function_response" in d]) == 0
    assert len([d for d in data if "function_call" in d]) == 1
    assert len([d for d in data if "raw_output" in d]) == 0
    mock_websocket.send.reset_mock()
    function_schemas_in_db[0]["mock_response"] = "13"

    # PARSING_FAILED_ERROR case

    system_prompt_with_format = """
You are a helpful assistant.

This is your output format. Keep the string between "[" and "]" as same as given below. You MUST FOLLOW this output format.
Output Format:
[response type=str]
(value here)
[/response]
    """
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_PROMPT_MODEL,
            "messages_for_run": [
                {"role": "system", "content": system_prompt_with_format, "step": 1},
                {
                    "role": "user",
                    "content": "Hello!",
                    "step": 2,
                },
            ],
            "model": "gpt-4-1106-preview",
            "parsing_type": ParsingType.SQUARE_BRACKET.value,
            "output_keys": ["respond"],
            "functions": ["get_current_weather"],
            "function_schemas": function_schemas_in_db,
        },
        ws=mock_websocket,
    )
    call_args_list = mock_websocket.send.call_args_list
    # print(call_args_list)
    data = [arg.args[0] for arg in call_args_list]

    assert len([d for d in data if d["status"] == "failed"]) == 1
    assert [d for d in data if d["status"] == "failed"][0][
        "error_type"
    ] == LocalTaskErrorType.PARSING_FAILED_ERROR.value
    assert len([d for d in data if d["status"] == "completed"]) == 0
    assert len([d for d in data if "function_response" in d]) == 0
    assert len([d for d in data if "function_call" in d]) == 0
    assert len([d for d in data if "raw_output" in d]) > 0
    assert len([d for d in data if "parsed_outputs" in d]) > 0

    mock_websocket.send.reset_mock()
    function_schemas_in_db[0]["mock_response"] = "13"

    # function not in code case, should use mock_response
    websocket_client._devapp.functions = {}
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_PROMPT_MODEL,
            "messages_for_run": [
                {
                    "role": "system",
                    "content": "You are a helpful assistant.",
                    "step": 1,
                },
                {
                    "role": "user",
                    "content": "What is the weather like in Boston?",
                    "step": 2,
                },
            ],
            "model": "gpt-3.5-turbo",
            "parsing_type": None,
            "output_keys": None,
            "functions": ["get_weather"],
            "function_schemas": function_schemas_in_db,
        },
        ws=mock_websocket,
    )
    call_args_list = mock_websocket.send.call_args_list
    print(call_args_list)
    data = [arg.args[0] for arg in call_args_list]

    assert len([d for d in data if d["status"] == "failed"]) == 0
    assert len([d for d in data if d["status"] == "completed"]) == 1

    assert len([d for d in data if "function_response" in d]) == 1
    assert (
        "FAKE RESPONSE"
        in [d for d in data if "function_response" in d][0]["function_response"][
            "response"
        ]
    )
    assert len([d for d in data if "function_call" in d]) == 1
    assert len([d for d in data if "raw_output" in d]) == 0



# import time
# import asyncio
# import typer
# import importlib
# import signal
# from typing import Dict, Any, List
# from playhouse.shortcuts import model_to_dict

# import webbrowser
# from rich import print
# from InquirerPy import inquirer
# from watchdog.observers import Observer

# from promptmodel import DevApp
# from promptmodel.apis.base import APIClient
# from promptmodel.constants import ENDPOINT_URL, WEB_CLIENT_URL
# from promptmodel.cli.commands.init import init as promptmodel_init
# from promptmodel.cli.utils import get_org, get_project
# from promptmodel.cli.signal_handler import dev_terminate_signal_handler
# from promptmodel.utils.config_utils import read_config, upsert_config
# from promptmodel.types.enums import ModelVersionStatus, ChangeLogAction
# from promptmodel.websocket import DevWebsocketClient, CodeReloadHandler
# from promptmodel.database.orm import initialize_db
# from promptmodel.database.models import *
# from promptmodel.database.crud import (
#     hide_function_model_not_in_code,
#     hide_chat_model_not_in_code,
#     update_samples,
#     update_function_model_uuid,
#     update_chat_model_uuid,
#     rename_function_model,
#     rename_chat_model,
# )


# def dev():
#     """Creates a new prompt development environment, and opens up DevApp in the browser."""
#     upsert_config({"initializing": True}, "connection")
#     signal.signal(signal.SIGINT, dev_terminate_signal_handler)
#     promptmodel_init(from_cli=False)

#     _devapp_filename, devapp_instance_name = "promptmodel_dev:app".split(":")

#     # Init local database & open
#     initialize_db()

#     config = read_config()

#     devapp_module = importlib.import_module(_devapp_filename)
#     devapp_instance: DevApp = getattr(devapp_module, devapp_instance_name)

#     if "name" not in config["connection"]:
#         org = get_org(config)
#         project = get_project(config=config, org=org)
#         validate_branch_name = lambda name: APIClient.execute(
#             method="GET", path="/check_connection_name", params={"name": name}
#         ).json()
#         branch_name = inquirer.text(
#             message="Enter a development branch name:",
#             validate=lambda x: " " not in x and validate_branch_name(x),
#             invalid_message="Branch name already exists or contains spaces.",
#         ).execute()

#         print("\nCreating local development branch...")
#         created_connection_row = APIClient.execute(
#             method="POST",
#             path="/create_connection",
#             params={"name": branch_name, "project_uuid": project["uuid"]},
#         )
#         upsert_config(
#             {
#                 "name": branch_name,
#                 "project": project,
#                 "org": org,
#                 "uuid": created_connection_row.json()["uuid"],
#                 "project_version": "0.0.0",
#             },
#             section="connection",
#         )

#         # connect
#         res = APIClient.execute(
#             method="POST",
#             path="/connect_cli_dev",
#             params={"project_uuid": project["uuid"], "branch_name": branch_name},
#         )
#         if res.status_code != 200:
#             print(f"Error: {res.json()['detail']}")
#             return

#         # fetch current project status
#         project_status = APIClient.execute(
#             method="GET",
#             path="/pull_project",
#             params={"project_uuid": project["uuid"]},
#         ).json()
#         # save project version
#         upsert_config(
#             {"project_version": project_status["project_version"]}, section="connection"
#         )

#         local_function_model_names = devapp_instance._get_function_model_name_list()
#         local_chat_model_names = devapp_instance._get_chat_model_name_list()

#         # save function_models
#         for function_model in project_status["function_models"]:
#             function_model["is_deployed"] = True
#             if function_model["name"] in local_function_model_names:
#                 function_model["used_in_code"] = True
#             else:
#                 function_model["used_in_code"] = False

#         FunctionModel.insert_many(project_status["function_models"]).execute()

#         # save function_model_versions
#         for version in project_status["function_model_versions"]:
#             version["status"] = ModelVersionStatus.CANDIDATE.value
#         FunctionModelVersion.insert_many(
#             project_status["function_model_versions"]
#         ).execute()
#         # save prompts
#         Prompt.insert_many(project_status["prompts"]).execute()
#         # save run_logs
#         RunLog.insert_many(project_status["run_logs"]).execute()

#         # create function_models from code in local DB
#         project_function_model_names = [
#             x["name"] for x in project_status["function_models"]
#         ]
#         only_in_local = list(
#             set(local_function_model_names) - set(project_function_model_names)
#         )
#         only_in_local_function_models = [
#             {"name": x, "project_uuid": project["uuid"]} for x in only_in_local
#         ]
#         FunctionModel.insert_many(only_in_local_function_models).execute()

#         # save chat_models
#         for chat_model in project_status["chat_models"]:
#             chat_model["is_deployed"] = True
#             if chat_model["name"] in local_chat_model_names:
#                 chat_model["used_in_code"] = True
#             else:
#                 chat_model["used_in_code"] = False

#         ChatModel.insert_many(project_status["chat_models"]).execute()

#         # save chat_model_versions
#         for version in project_status["chat_model_versions"]:
#             version["status"] = ModelVersionStatus.CANDIDATE.value
#         ChatModelVersion.insert_many(project_status["chat_model_versions"]).execute()

#         # create_chat_models from code in local DB
#         project_chat_model_names = [x["name"] for x in project_status["chat_models"]]
#         only_in_local_names = list(
#             set(local_chat_model_names) - set(project_chat_model_names)
#         )
#         only_in_local_chat_models = [
#             {"name": x, "project_uuid": project["uuid"]} for x in only_in_local_names
#         ]
#         ChatModel.insert_many(only_in_local_chat_models).execute()

#     else:
#         org = config["connection"]["org"]
#         project = config["connection"]["project"]
#         branch_name = config["connection"]["name"]

#         res = APIClient.execute(
#             method="POST",
#             path="/connect_cli_dev",
#             params={"project_uuid": project["uuid"], "branch_name": branch_name},
#         )
#         if res.status_code != 200:
#             print(f"Error: {res.json()['detail']}")
#             return

#         # Get changelog & update DB
#         # fetch current project status
#         project_status = APIClient.execute(
#             method="GET",
#             path="/pull_project",
#             params={"project_uuid": project["uuid"]},
#         ).json()

#         changelogs = APIClient.execute(
#             method="GET",
#             path="/get_changelog",
#             params={
#                 "project_uuid": project["uuid"],
#                 "local_project_version": config["connection"]["project_version"],
#                 "levels": [1, 2],
#             },
#         ).json()
#         local_code_function_model_name_list = (
#             devapp_instance._get_function_model_name_list()
#         )
#         local_code_chat_model_name_list = devapp_instance._get_chat_model_name_list()

#         res = update_by_changelog(
#             changelogs,
#             project_status,
#             local_code_function_model_name_list,
#             local_code_chat_model_name_list,
#         )

#         if res is False:
#             print("Update Dev Stopped.")
#             upsert_config({"online": False}, section="connection")
#             return

#         # Make function_model.used_in_code=False which is not in local code
#         hide_function_model_not_in_code(local_code_function_model_name_list)
#         # Make chat_model.used_in_code=False which is not in local code
#         hide_chat_model_not_in_code(local_code_chat_model_name_list)

#         # Save new function_model in code to local DB
#         local_db_function_model_names = [
#             x["name"]
#             for x in [
#                 model_to_dict(x, recurse=False) for x in list(FunctionModel.select())
#             ]
#         ]
#         only_in_local = list(
#             set(local_code_function_model_name_list) - set(local_db_function_model_names)
#         )
#         only_in_local_function_models = [
#             {"name": x, "project_uuid": project["uuid"]} for x in only_in_local
#         ]
#         FunctionModel.insert_many(only_in_local_function_models).execute()

#         # Save new chat_model in code to local DB
#         local_db_chat_model_names = [
#             x["name"]
#             for x in [model_to_dict(x, recurse=False) for x in list(ChatModel.select())]
#         ]
#         only_in_local = list(
#             set(local_code_chat_model_name_list) - set(local_db_chat_model_names)
#         )
#         only_in_local_chat_models = [
#             {"name": x, "project_uuid": project["uuid"]} for x in only_in_local
#         ]
#         ChatModel.insert_many(only_in_local_chat_models).execute()

#     dev_url = f"{WEB_CLIENT_URL}/org/{org['slug']}/projects/{project['uuid']}/dev/{branch_name}"

#     # Open websocket connection to backend server
#     dev_websocket_client = DevWebsocketClient(_devapp=devapp_instance)

#     import threading

#     reloader_thread = threading.Thread(
#         target=start_code_reloader,
#         args=(_devapp_filename, devapp_instance_name, dev_websocket_client),
#     )
#     reloader_thread.daemon = True  # Set the thread as a daemon
#     reloader_thread.start()

#     print(
#         f"\nOpening [violet]Promptmodel[/violet] prompt engineering environment with the following configuration:\n"
#     )
#     print(f"📌 Organization: [blue]{org['name']}[/blue]")
#     print(f"📌 Project: [blue]{project['name']}[/blue]")
#     print(f"📌 Local development branch: [green][bold]{branch_name}[/bold][/green]")
#     print(
#         f"\nIf browser doesn't open automatically, please visit [link={dev_url}]{dev_url}[/link]"
#     )
#     webbrowser.open(dev_url)

#     upsert_config({"online": True, "initializing": False}, section="connection")
#     # save samples to local DB
#     update_samples(devapp_instance.samples)

#     # Open Websocket
#     asyncio.run(
#         dev_websocket_client.connect_to_gateway(
#             project_uuid=project["uuid"],
#             connection_name=branch_name,
#             cli_access_header=APIClient._get_headers(),
#         )
#     )


# app = typer.Typer(invoke_without_command=True, callback=dev)


# def start_code_reloader(_devapp_filename, devapp_instance_name, dev_websocket_client):
#     event_handler = CodeReloadHandler(
#         _devapp_filename, devapp_instance_name, dev_websocket_client
#     )
#     observer = Observer()
#     observer.schedule(event_handler, path=".", recursive=True)
#     observer.start()

#     try:
#         while True:
#             time.sleep(1)
#     except KeyboardInterrupt:
#         observer.stop()
#     observer.join()


# def update_by_changelog(
#     changelogs: List[Dict],
#     project_status: dict,
#     local_code_function_model_name_list: List[str],
#     local_code_chat_model_name_list: List[str],
# ):
#     """Update Local DB by changelog"""
#     local_db_function_model_list: List = [
#         model_to_dict(x, recurse=False) for x in list(FunctionModel.select())
#     ]  # {"name", "uuid"}
#     local_db_chat_model_list: List = [
#         model_to_dict(x, recurse=False) for x in list(ChatModel.select())
#     ]  # {"name", "uuid"}

#     for changelog in changelogs:
#         level: int = changelog["level"]
#         logs = changelog["logs"]
#         if level == 1:
#             for log in logs:
#                 subject = log["subject"]
#                 action: str = log["action"]
#                 if subject == "function_model":
#                     (
#                         is_success,
#                         local_db_function_model_list,
#                     ) = update_function_model_changelog(
#                         action=action,
#                         project_status=project_status,
#                         uuid_list=log["identifiers"],
#                         local_db_function_model_list=local_db_function_model_list,
#                         local_code_function_model_name_list=local_code_function_model_name_list,
#                     )

#                     if not is_success:
#                         return False

#                 elif subject == "function_model_version":
#                     local_db_function_model_list = update_function_model_version_changelog(
#                         action=action,
#                         project_status=project_status,
#                         uuid_list=log["identifiers"],
#                         local_db_function_model_list=local_db_function_model_list,
#                         local_code_function_model_name_list=local_code_function_model_name_list,
#                     )

#                 elif subject == "chat_model":
#                     is_success, local_db_chat_model_list = update_chat_model_changelog(
#                         action=action,
#                         project_status=project_status,
#                         uuid_list=log["identifiers"],
#                         local_db_chat_model_list=local_db_chat_model_list,
#                         local_code_chat_model_name_list=local_code_chat_model_name_list,
#                     )

#                     if not is_success:
#                         return False

#                 elif subject == "chat_model_version":
#                     local_db_chat_model_list = update_chat_model_version_changelog(
#                         action=action,
#                         project_status=project_status,
#                         uuid_list=log["identifiers"],
#                         local_db_chat_model_list=local_db_chat_model_list,
#                         local_code_chat_model_name_list=local_code_chat_model_name_list,
#                     )

#                 else:
#                     pass
#             previous_version_levels = changelog["previous_version"].split(".")
#             current_version_levels = [
#                 str(int(previous_version_levels[0]) + 1),
#                 "0",
#                 "0",
#             ]
#             current_version = ".".join(current_version_levels)
#         elif level == 2:
#             for log in logs:
#                 subject = log["subject"]
#                 action: str = log["action"]
#                 uuid_list: list = log["identifiers"]
#                 if subject == "function_model_version":
#                     local_db_function_model_list = update_function_model_version_changelog(
#                         action=action,
#                         project_status=project_status,
#                         uuid_list=log["identifiers"],
#                         local_db_function_model_list=local_db_function_model_list,
#                         local_code_function_model_name_list=local_code_function_model_name_list,
#                     )

#                 elif subject == "chat_model_version":
#                     local_db_chat_model_list = update_chat_model_version_changelog(
#                         action=action,
#                         project_status=project_status,
#                         uuid_list=log["identifiers"],
#                         local_db_chat_model_list=local_db_chat_model_list,
#                         local_code_chat_model_name_list=local_code_chat_model_name_list,
#                     )
#                 else:
#                     pass
#             previous_version_levels = changelog["previous_version"].split(".")
#             current_version_levels = [
#                 previous_version_levels[0],
#                 str(int(previous_version_levels[1]) + 1),
#                 "0",
#             ]
#             current_version = ".".join(current_version_levels)
#         else:
#             previous_version_levels = changelog["previous_version"].split(".")
#             current_version_levels = [
#                 previous_version_levels[0],
#                 previous_version_levels[1],
#                 str(int(previous_version_levels[2]) + 1),
#             ]
#             current_version = ".".join(current_version_levels)

#         upsert_config({"project_version": current_version}, section="connection")
#     return True


# def update_function_model_changelog(
#     action: ChangeLogAction,
#     project_status: dict,
#     uuid_list: List[str],
#     local_db_function_model_list: List[Dict],
#     local_code_function_model_name_list: List[str],
# ):
#     if action == ChangeLogAction.ADD.value:
#         function_model_list = [
#             x for x in project_status["function_models"] if x["uuid"] in uuid_list
#         ]
#         for function_model in function_model_list:
#             local_db_function_model_name_list = [
#                 x["name"] for x in local_db_function_model_list
#             ]

#             if function_model["name"] not in local_db_function_model_name_list:
#                 # IF function_model not in Local DB
#                 if function_model["name"] in local_code_function_model_name_list:
#                     # IF function_model in Local Code
#                     function_model["used_in_code"] = True
#                     function_model["is_deployed"] = True
#                 else:
#                     function_model["used_in_code"] = False
#                     function_model["is_deployed"] = True
#                 FunctionModel.create(**function_model)
#             else:
#                 local_db_function_model = [
#                     x
#                     for x in local_db_function_model_list
#                     if x["name"] == function_model["name"]
#                 ][0]
#                 if local_db_function_model["is_deployed"] is False:
#                     print(
#                         "Creation of promptmodel with identical name was detected in local & deployment."
#                     )
#                     check_same = inquirer.confirm(
#                         message="Are they same promptmodel? [y/n]", default=False
#                     ).execute()
#                     # check_same = input()
#                     # if check_same == "y":
#                     #     check_same = True
#                     # else:
#                     #     check_same = False

#                     if not check_same:
#                         # rename & change name in Local DB
#                         print(
#                             f"Please rename local promptmodel {function_model['name']} to continue"
#                         )
#                         validate_new_promptmodel_name = (
#                             lambda name: name not in local_code_function_model_name_list
#                             and name not in local_db_function_model_name_list
#                         )
#                         new_model_name = inquirer.text(
#                             message="Enter the new promptmodel name:",
#                             validate=lambda x: validate_new_promptmodel_name(x),
#                             invalid_message="promptmodel name already exists.",
#                         ).execute()
#                         # new_model_name = input()
#                         rename_function_model(
#                             local_db_function_model["uuid"], new_model_name
#                         )

#                         print("We changed the name of promptmodel in local DB.")
#                         print(
#                             f"Please change the name of promptmodel '{function_model['name']}' in your project code and restart."
#                         )
#                         # dev 꺼버리고, 수정하고 다시 키라고 명령.
#                         return False, local_db_function_model_list
#                     update_function_model_uuid(
#                         local_db_function_model["uuid"], function_model["uuid"]
#                     )
#                     local_db_function_model_list: list = [
#                         model_to_dict(x, recurse=False)
#                         for x in list(FunctionModel.select())
#                     ]
#     else:
#         # TODO: add code DELETE, CHANGE, FIX later
#         pass
#     return True, local_db_function_model_list


# def update_function_model_version_changelog(
#     action: ChangeLogAction,
#     project_status: dict,
#     uuid_list: List[str],
#     local_db_function_model_list: List[Dict],
#     local_code_function_model_name_list: List[str],
# ) -> List[Dict[str, Any]]:
#     if action == ChangeLogAction.ADD.value:
#         # find function_model_version in project_status['function_model_versions'] where uuid in uuid_list
#         function_model_version_list_in_changelog = [
#             x for x in project_status["function_model_versions"] if x["uuid"] in uuid_list
#         ]

#         # check if function_model_version['uuid'] is in local_db_function_model_list
#         local_db_version_uuid_list = [
#             str(x.uuid) for x in list(FunctionModelVersion.select())
#         ]
#         version_list_to_update = [
#             x
#             for x in function_model_version_list_in_changelog
#             if x["uuid"] not in local_db_version_uuid_list
#         ]
#         version_uuid_list_to_update = [x["uuid"] for x in version_list_to_update]

#         # find prompts and run_logs to update
#         prompts_to_update = [
#             x
#             for x in project_status["prompts"]
#             if x["version_uuid"] in version_uuid_list_to_update
#         ]
#         run_logs_to_update = [
#             x
#             for x in project_status["run_logs"]
#             if x["version_uuid"] in version_uuid_list_to_update
#         ]

#         for function_model_version in version_list_to_update:
#             function_model_version["status"] = ModelVersionStatus.CANDIDATE.value

#         FunctionModelVersion.insert_many(version_list_to_update).execute()
#         Prompt.insert_many(prompts_to_update).execute()
#         RunLog.insert_many(run_logs_to_update).execute()

#         return local_db_function_model_list
#     else:
#         pass


# def update_chat_model_changelog(
#     action: ChangeLogAction,
#     project_status: dict,
#     uuid_list: List[str],
#     local_db_chat_model_list: List[Dict],
#     local_code_chat_model_name_list: List[str],
# ):
#     if action == ChangeLogAction.ADD.value:
#         chat_model_list = [
#             x for x in project_status["chat_models"] if x["uuid"] in uuid_list
#         ]
#         for chat_model in chat_model_list:
#             local_db_chat_model_name_list = [
#                 x["name"] for x in local_db_chat_model_list
#             ]

#             if chat_model["name"] not in local_db_chat_model_name_list:
#                 # IF chat_model not in Local DB
#                 if chat_model["name"] in local_code_chat_model_name_list:
#                     # IF chat_model in Local Code
#                     chat_model["used_in_code"] = True
#                     chat_model["is_deployed"] = True
#                 else:
#                     chat_model["used_in_code"] = False
#                     chat_model["is_deployed"] = True
#                 ChatModel.create(**chat_model)
#             else:
#                 local_db_chat_model = [
#                     x
#                     for x in local_db_chat_model_list
#                     if x["name"] == chat_model["name"]
#                 ][0]
#                 if local_db_chat_model["is_deployed"] is False:
#                     print(
#                         "Creation of chatmodel with identical name was detected in local & deployment."
#                     )
#                     check_same = inquirer.confirm(
#                         message="Are they same chatmodel? [y/n]", default=False
#                     ).execute()
#                     # check_same = input()
#                     # if check_same == "y":
#                     #     check_same = True
#                     # else:
#                     #     check_same = False

#                     if not check_same:
#                         # rename & change name in Local DB
#                         print(
#                             f"Please rename local chatmodel {chat_model['name']} to continue"
#                         )
#                         validate_new_chatmodel_name = (
#                             lambda name: name not in local_code_chat_model_name_list
#                             and name not in local_db_chat_model_name_list
#                         )
#                         new_model_name = inquirer.text(
#                             message="Enter the new chatmodel name:",
#                             validate=lambda x: validate_new_chatmodel_name(x),
#                             invalid_message="chatmodel name already exists.",
#                         ).execute()
#                         # new_model_name = input()
#                         rename_chat_model(local_db_chat_model["uuid"], new_model_name)

#                         print("We changed the name of chatmodel in local DB.")
#                         print(
#                             f"Please change the name of chatmodel '{chat_model['name']}' in your project code and restart."
#                         )
#                         # dev 꺼버리고, 수정하고 다시 키라고 명령.
#                         return False, local_db_chat_model_list
#                     update_chat_model_uuid(
#                         local_db_chat_model["uuid"], chat_model["uuid"]
#                     )
#                     local_db_chat_model_list: list = [
#                         model_to_dict(x, recurse=False)
#                         for x in list(ChatModel.select())
#                     ]
#     else:
#         # TODO: add code DELETE, CHANGE, FIX later
#         pass
#     return True, local_db_chat_model_list


# def update_chat_model_version_changelog(
#     action: ChangeLogAction,
#     project_status: dict,
#     uuid_list: List[str],
#     local_db_chat_model_list: List[Dict],
#     local_code_chat_model_name_list: List[str],
# ) -> List[Dict[str, Any]]:
#     if action == ChangeLogAction.ADD.value:
#         # find chat_model_version in project_status['chat_model_versions'] where uuid in uuid_list
#         chat_model_version_list_in_changelog = [
#             x for x in project_status["chat_model_versions"] if x["uuid"] in uuid_list
#         ]

#         # check if chat_model_version['uuid'] is in local_db_chat_model_list
#         local_db_version_uuid_list = [
#             str(x.uuid) for x in list(ChatModelVersion.select())
#         ]
#         version_list_to_update = [
#             x
#             for x in chat_model_version_list_in_changelog
#             if x["uuid"] not in local_db_version_uuid_list
#         ]

#         for chat_model_version in version_list_to_update:
#             chat_model_version["status"] = ModelVersionStatus.CANDIDATE.value

#         ChatModelVersion.insert_many(version_list_to_update).execute()

#         return local_db_chat_model_list
#     else:
#         pass



import signal
import sys
from promptmodel.utils.config_utils import upsert_config, read_config


def dev_terminate_signal_handler(sig, frame):
    config = read_config()
    print("\nTerminating...")
    if "connection" in config:
        upsert_config({"online": False}, section="connection")
        upsert_config({"initializing": False}, "connection")
    sys.exit(0)



import datetime
import json

from uuid import uuid4
from peewee import (
    CharField,
    DateTimeField,
    IntegerField,
    ForeignKeyField,
    BooleanField,
    UUIDField,
    TextField,
    AutoField,
    FloatField,
    Check,
)

from promptmodel.database.config import BaseModel
from promptmodel.database.models_chat import *
from promptmodel.types.enums import ParsingType


class JSONField(TextField):
    def db_value(self, value):
        return json.dumps(value)

    def python_value(self, value):
        return json.loads(value)


class DeployedFunctionModel(BaseModel):
    uuid = UUIDField(unique=True, default=uuid4)
    name = CharField()


class DeployedFunctionModelVersion(BaseModel):
    uuid = UUIDField(unique=True, default=uuid4)
    version = IntegerField(null=False)
    from_version = IntegerField(null=True)
    function_model_uuid = ForeignKeyField(
        DeployedFunctionModel,
        field=DeployedFunctionModel.uuid,
        backref="versions",
        on_delete="CASCADE",
    )
    model = CharField()
    is_published = BooleanField(default=False)
    is_ab_test = BooleanField(default=False)
    ratio = FloatField(null=True)
    parsing_type = CharField(
        null=True,
        default=None,
        constraints=[
            Check(
                f"parsing_type IN ('{ParsingType.COLON.value}', '{ParsingType.SQUARE_BRACKET.value}', '{ParsingType.DOUBLE_SQUARE_BRACKET.value}')"
            )
        ],
    )
    output_keys = JSONField(null=True, default=None)
    functions = JSONField(default=[])


class DeployedPrompt(BaseModel):
    id = AutoField()
    version_uuid = ForeignKeyField(
        DeployedFunctionModelVersion,
        field=DeployedFunctionModelVersion.uuid,
        backref="prompts",
        on_delete="CASCADE",
    )
    role = CharField()
    step = IntegerField()
    content = TextField()



"""LLM for Development TestRun"""
import re
from typing import Any, AsyncGenerator, List, Dict, Optional
from pydantic import BaseModel
from dotenv import load_dotenv
from litellm import acompletion, get_max_tokens

from promptmodel.types.enums import ParsingType, get_pattern_by_type
from promptmodel.utils import logger
from promptmodel.utils.output_utils import convert_str_to_type, update_dict
from promptmodel.utils.token_counting import (
    num_tokens_for_messages_for_each,
    num_tokens_from_functions_input,
)
from promptmodel.types.response import LLMStreamResponse, ModelResponse


load_dotenv()


class OpenAIMessage(BaseModel):
    role: Optional[str] = None
    content: Optional[str] = ""
    function_call: Optional[Dict[str, Any]] = None
    name: Optional[str] = None


class LLMDev:
    def __init__(self):
        self._model: str

    def __validate_openai_messages(
        self, messages: List[Dict[str, Any]]
    ) -> List[OpenAIMessage]:
        """Validate and convert list of dictionaries to list of OpenAIMessage."""
        res = []
        for message in messages:
            res.append(OpenAIMessage(**message))
        return res

    async def dev_run(
        self,
        messages: List[Dict[str, Any]],
        parsing_type: Optional[ParsingType] = None,
        functions: Optional[List[Any]] = None,
        model: Optional[str] = None,
        **kwargs,
    ) -> AsyncGenerator[Any, None]:
        """Parse & stream output from openai chat completion."""
        _model = model or self._model
        raw_output = ""
        if functions == []:
            functions = None
        response: AsyncGenerator[ModelResponse, None] = await acompletion(
            model=_model,
            messages=[
                message.model_dump(exclude_none=True)
                for message in self.__validate_openai_messages(messages)
            ],
            stream=True,
            functions=functions,
            **kwargs,
        )
        function_call = {"name": "", "arguments": ""}
        finish_reason_function_call = False
        async for chunk in response:
            if getattr(chunk.choices[0].delta, "content", None) is not None:
                stream_value = chunk.choices[0].delta.content
                raw_output += stream_value  # append raw output
                yield LLMStreamResponse(raw_output=stream_value)  # return raw output

            if getattr(chunk.choices[0].delta, "function_call", None) is not None:
                for key, value in (
                    chunk.choices[0].delta.function_call.model_dump().items()
                ):
                    if value is not None:
                        function_call[key] += value

            if chunk.choices[0].finish_reason == "function_call":
                finish_reason_function_call = True
                yield LLMStreamResponse(function_call=function_call)

        # parsing
        if parsing_type and not finish_reason_function_call:
            parsing_pattern: Dict[str, str] = get_pattern_by_type(parsing_type)
            whole_pattern = parsing_pattern["whole"]
            parsed_results = re.findall(whole_pattern, raw_output, flags=re.DOTALL)
            for parsed_result in parsed_results:
                key = parsed_result[0]
                type_str = parsed_result[1]
                value = convert_str_to_type(parsed_result[2], type_str)
                yield LLMStreamResponse(parsed_outputs={key: value})

    async def dev_chat(
        self,
        messages: List[Dict[str, Any]],
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        model: Optional[str] = None,
        **kwargs,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        """Parse & stream output from openai chat completion."""
        _model = model or self._model
        raw_output = ""
        if functions == []:
            functions = None

        if model != "HCX-002":
            # Truncate the output if it is too long
            # truncate messages to make length <= model's max length
            token_per_functions = num_tokens_from_functions_input(
                functions=functions, model=model
            )
            model_max_tokens = get_max_tokens(model=model)
            token_per_messages = num_tokens_for_messages_for_each(messages, model)
            token_limit_exceeded = (
                sum(token_per_messages) + token_per_functions
            ) - model_max_tokens
            if token_limit_exceeded > 0:
                while token_limit_exceeded > 0:
                    # erase the second oldest message (first one is system prompt, so it should not be erased)
                    if len(messages) == 1:
                        # if there is only one message, Error cannot be solved. Just call LLM and get error response
                        break
                    token_limit_exceeded -= token_per_messages[1]
                    del messages[1]
                    del token_per_messages[1]

        args = dict(
            model=_model,
            messages=[
                message.model_dump(exclude_none=True)
                for message in self.__validate_openai_messages(messages)
            ],
            functions=functions,
            tools=tools,
        )

        is_stream_unsupported = model in ["HCX-002"]
        if not is_stream_unsupported:
            args["stream"] = True
        response: AsyncGenerator[ModelResponse, None] = await acompletion(**args, **kwargs)
        if is_stream_unsupported:
            yield LLMStreamResponse(raw_output=response.choices[0].message.content)
        else:
            async for chunk in response:
                yield_api_response_with_fc = False
                logger.debug(chunk)
                if getattr(chunk.choices[0].delta, "function_call", None) is not None:
                    yield LLMStreamResponse(
                        api_response=chunk,
                        function_call=chunk.choices[0].delta.function_call,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "tool_calls", None) is not None:
                    yield LLMStreamResponse(
                        api_response=chunk,
                        tool_calls=chunk.choices[0].delta.tool_calls,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "content", None) is not None:
                    raw_output += chunk.choices[0].delta.content
                    yield LLMStreamResponse(
                        api_response=chunk if not yield_api_response_with_fc else None,
                        raw_output=chunk.choices[0].delta.content,
                    )



import os
from typing import Any, Dict
from cryptography.fernet import Fernet
import secrets


def generate_api_key():
    """Generate a random token with 32 bytes"""
    token = secrets.token_hex(32)
    return token


def generate_crypto_key():
    """
    Generates a key and save it into a file
    """
    key = Fernet.generate_key()
    with open("key.key", "wb") as key_file:
        key_file.write(key)


def load_crypto_key():
    """
    Loads the key named `key.key` from the directory.
    Generates a new key if it does not exist.
    """
    if not os.path.exists("key.key"):
        generate_crypto_key()
    return open("key.key", "rb").read()


def encrypt_message(message: str) -> bytes:
    """
    Encrypts the message
    """
    key = load_crypto_key()
    f = Fernet(key)
    message = message.encode()
    encrypted_message = f.encrypt(message)
    return encrypted_message


def decrypt_message(encrypted_message: bytes) -> str:
    """
    Decrypts the message
    """
    key = load_crypto_key()
    f = Fernet(key)
    decrypted_message = f.decrypt(encrypted_message)
    return decrypted_message.decode()



#!/Users/jypark/Desktop/weavel/promptmodel-python/venv/bin/python3.8

# $Id: rst2xml.py 9115 2022-07-28 17:06:24Z milde $
# Author: David Goodger <goodger@python.org>
# Copyright: This module has been placed in the public domain.

"""
A minimal front end to the Docutils Publisher, producing Docutils XML.
"""

try:
    import locale
    locale.setlocale(locale.LC_ALL, '')
except Exception:
    pass

from docutils.core import publish_cmdline, default_description


description = ('Generates Docutils-native XML from standalone '
               'reStructuredText sources.  ' + default_description)

publish_cmdline(writer_name='xml', description=description)



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any, Optional
from litellm import ModelResponse

from promptmodel.llms.llm import LLM, ParseResult
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.types.response import LLMResponse, LLMStreamResponse, ChatModelConfig
from promptmodel.types.enums import ParsingType
from promptmodel import ChatModel, DevClient
from promptmodel.dev_app import ChatModelInterface, ChatModelInterface
from promptmodel.chat_model import RegisteringMeta

client = DevClient()


def test_find_client(
    mocker,
    mock_fetch_chat_model,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    fetch_chat_model = mocker.patch(
        "promptmodel.chat_model.LLMProxy.fetch_chat_model",
        mock_fetch_chat_model,
    )

    mocker.patch(
        "promptmodel.chat_model.LLMProxy._async_make_session_cloud",
        mock_async_chat_log_to_cloud,
    )
    pm = ChatModel("test")
    assert client.chat_models == [ChatModelInterface(name="test")]


def test_get_config(
    mocker,
    mock_fetch_chat_model,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    fetch_chat_model = mocker.patch(
        "promptmodel.chat_model.LLMProxy.fetch_chat_model", mock_fetch_chat_model
    )

    mocker.patch(
        "promptmodel.chat_model.LLMProxy._async_make_session_cloud",
        mock_async_chat_log_to_cloud,
    )
    # mock registering_meta
    mocker.patch("promptmodel.chat_model.RegisteringMeta", MagicMock())
    chat_model = ChatModel("test")
    assert len(client.chat_models) == 1
    config: ChatModelConfig = chat_model.get_config()
    assert config.system_prompt == "You are a helpful assistant."


def test_add_messages(
    mocker,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    pass


def test_run(
    mocker,
    mock_fetch_chat_model: AsyncMock,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    fetch_chat_model = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_chat_model", mock_fetch_chat_model
    )

    async_chat_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_chat_log_to_cloud",
        mock_async_chat_log_to_cloud,
    )

    mocker.patch("promptmodel.chat_model.RegisteringMeta", MagicMock())
    chat_model = ChatModel("test", session_uuid="testuuid")

    res: LLMResponse = chat_model.run()
    print(res.api_response.model_dump())

    fetch_chat_model.assert_called_once()
    async_chat_log_to_cloud.assert_called_once()

    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs is None

    fetch_chat_model.reset_mock()
    async_chat_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )
    res: LLMResponse = chat_model.run()
    print(res)
    fetch_chat_model.assert_not_called()
    async_chat_log_to_cloud.assert_not_called()


@pytest.mark.asyncio
async def test_arun(
    mocker,
    mock_fetch_chat_model: AsyncMock,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    fetch_chat_model = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_chat_model", mock_fetch_chat_model
    )

    async_chat_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_chat_log_to_cloud",
        mock_async_chat_log_to_cloud,
    )
    mocker.patch("promptmodel.chat_model.RegisteringMeta", MagicMock())
    chat_model = ChatModel("test", session_uuid="testuuid")

    res: LLMResponse = await chat_model.arun()
    print(res.api_response.model_dump())

    fetch_chat_model.assert_called_once()
    async_chat_log_to_cloud.assert_called_once()

    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs is None

    fetch_chat_model.reset_mock()
    async_chat_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )
    res: LLMResponse = await chat_model.arun()
    print(res)
    fetch_chat_model.assert_not_called()
    async_chat_log_to_cloud.assert_not_called()


def test_stream(
    mocker,
    mock_fetch_chat_model: AsyncMock,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    fetch_chat_model = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_chat_model", mock_fetch_chat_model
    )

    async_chat_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_chat_log_to_cloud",
        mock_async_chat_log_to_cloud,
    )
    mocker.patch("promptmodel.chat_model.RegisteringMeta", MagicMock())
    chat_model = ChatModel("test", session_uuid="testuuid")

    res: Generator[LLMStreamResponse, None, None] = chat_model.run(stream=True)
    chunks: List[LLMStreamResponse] = []
    for chunk in res:
        chunks.append(chunk)

    fetch_chat_model.assert_called_once()
    async_chat_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0

    fetch_chat_model.reset_mock()
    async_chat_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )
    res: LLMResponse = chat_model.run(stream=True)
    print(res)
    fetch_chat_model.assert_not_called()
    async_chat_log_to_cloud.assert_not_called()


@pytest.mark.asyncio
async def test_astream(
    mocker,
    mock_fetch_chat_model: AsyncMock,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    fetch_chat_model = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_chat_model", mock_fetch_chat_model
    )

    async_chat_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_chat_log_to_cloud",
        mock_async_chat_log_to_cloud,
    )

    mocker.patch("promptmodel.chat_model.RegisteringMeta", MagicMock())
    chat_model = ChatModel("test", session_uuid="testuuid")

    res: AsyncGenerator[LLMStreamResponse, None] = await chat_model.arun(stream=True)

    chunks: List[LLMStreamResponse] = []
    async for chunk in res:
        chunks.append(chunk)

    fetch_chat_model.assert_called_once()
    async_chat_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0

    fetch_chat_model.reset_mock()
    async_chat_log_to_cloud.reset_mock()
    mocker.patch(
        "promptmodel.utils.config_utils.read_config",
        return_value={"connection": {"initializing": True}},
    )

    res: LLMResponse = await chat_model.arun(stream=True)
    print(res)
    fetch_chat_model.assert_not_called()
    async_chat_log_to_cloud.assert_not_called()



import os
from dotenv import load_dotenv

load_dotenv()
testmode: str = os.environ.get("TESTMODE", "false")

if testmode == "true":
    ENDPOINT_URL = (
        os.environ.get(
            "TESTMODE_PROMPTMODEL_BACKEND_PUBLIC_URL", "http://localhost:8000"
        )
        + "/api/cli"
    )
    WEB_CLIENT_URL = os.environ.get(
        "TESTMODE_PROMPTMODEL_FRONTEND_PUBLIC_URL", "http://localhost:3000"
    )
    GRANT_ACCESS_URL = WEB_CLIENT_URL + "/cli/grant-access"
else:
    ENDPOINT_URL = (
        os.environ.get(
            "PROMPTMODEL_BACKEND_PUBLIC_URL", "https://promptmodel.up.railway.app"
        )
        + "/api/cli"
    )

    WEB_CLIENT_URL = os.environ.get(
        "PROMPTMODEL_FRONTEND_PUBLIC_URL", "https://app.promptmodel.run"
    )
    GRANT_ACCESS_URL = WEB_CLIENT_URL + "/cli/grant-access"

PROMPTMODEL_DEV_FILENAME = os.path.join(os.getcwd(), "promptmodel_dev.py")
PROMPTMODEL_DEV_STARTER_FILENAME = "STARTER.py"



import os
import nest_asyncio
import threading
import asyncio
import atexit
import time

from typing import Optional, Dict, Any
from datetime import datetime

from promptmodel.utils.config_utils import upsert_config, read_config
from promptmodel.utils import logger
from promptmodel.database.orm import initialize_db
from promptmodel.database.crud import update_deployed_cache
from promptmodel.apis.base import AsyncAPIClient
from promptmodel.types.enums import InstanceType


def init(use_cache: Optional[bool] = True, mask_inputs: Optional[bool] = False):
    nest_asyncio.apply()

    config = read_config()
    if (
        config
        and "connection" in config
        and (
            (
                "online" in config["connection"]
                and config["connection"]["online"] == True
            )
            or (
                "initializing" in config["connection"]
                and config["connection"]["initializing"] == True
            )
        )
    ):
        cache_manager = None
    else:
        if use_cache:
            upsert_config({"use_cache": True}, section="project")
            cache_manager = CacheManager()
        else:
            upsert_config({"use_cache": False}, section="project")
            cache_manager = None
            initialize_db()  # init db for local usage

        if mask_inputs is True:
            upsert_config({"mask_inputs": True}, section="project")
        else:
            upsert_config({"mask_inputs": False}, section="project")


class CacheManager:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls):
        with cls._lock:
            if cls._instance is None:
                instance = super(CacheManager, cls).__new__(cls)
                instance.last_update_time = 0  # to manage update frequency
                instance.update_interval = 60 * 60 * 6  # seconds, 6 hours
                instance.program_alive = True
                instance.background_tasks = []
                initialize_db()
                atexit.register(instance._terminate)
                asyncio.run(instance.update_cache())  # updae cache first synchronously
                instance.cache_thread = threading.Thread(
                    target=instance._run_cache_loop
                )
                instance.cache_thread.daemon = True
                instance.cache_thread.start()
                cls._instance = instance
        return cls._instance

    def cache_update_background_task(self, config):
        asyncio.run(update_deployed_db(config))

    def _run_cache_loop(self):
        asyncio.run(self._update_cache_periodically())

    async def _update_cache_periodically(self):
        while True:
            await asyncio.sleep(self.update_interval)  # Non-blocking sleep
            await self.update_cache()

    async def update_cache(self):
        # Current time
        current_time = time.time()
        config = read_config()

        if not config:
            upsert_config({"version": 0}, section="project")
            config = {"project": {"version": 0}}
        if "project" not in config:
            upsert_config({"version": 0}, section="project")
            config = {"project": {"version": 0}}

        if "version" not in config["project"]:
            upsert_config({"version": 0}, section="project")
            config = {"project": {"version": 0}}

        # Check if we need to update the cache
        if current_time - self.last_update_time > self.update_interval:
            # Update cache logic
            try:
                await update_deployed_db(config)
            except:
                # try once more
                await update_deployed_db(config)
            # Update the last update time
            self.last_update_time = current_time

    def _terminate(self):
        self.program_alive = False

    # async def cleanup_background_tasks(self):
    #     for task in self.background_tasks:
    #         if not task.done():
    #             task.cancel()
    #         try:
    #             await task
    #         except asyncio.CancelledError:
    #             pass  # 작업이 취소됨


async def update_deployed_db(config):
    if "project" not in config or "version" not in config["project"]:
        cached_project_version = 0
    else:
        cached_project_version = int(config["project"]["version"])
    try:
        res = await AsyncAPIClient.execute(
            method="GET",
            path="/check_update",
            params={"cached_version": cached_project_version},
            use_cli_key=False,
        )
        res = res.json()
        if res["need_update"]:
            # update local DB with res['project_status']
            project_status = res["project_status"]
            await update_deployed_cache(project_status)
            upsert_config({"version": res["version"]}, section="project")
        else:
            upsert_config({"version": res["version"]}, section="project")
    except Exception as exception:
        logger.error(f"Deployment cache update error: {exception}")



from promptmodel.apis.base import APIClient
import typer
from InquirerPy import inquirer

from promptmodel.utils.config_utils import upsert_config


def configure():
    """Saves user's default organization and project."""

    orgs = APIClient.execute(method="GET", path="/organizations").json()
    choices = [
        {
            "key": org["name"],
            "name": org["name"],
            "value": org,
        }
        for org in orgs
    ]
    org = inquirer.select(
        message="Select default organization:", choices=choices
    ).execute()

    projects = APIClient.execute(
        method="GET",
        path="/projects",
        params={"organization_id": org["organization_id"]},
    ).json()
    choices = [
        {
            "key": project["name"],
            "name": project["name"],
            "value": project,
        }
        for project in projects
    ]
    project = inquirer.select(
        message="Select default project:", choices=choices
    ).execute()

    upsert_config({"org": org, "project": project}, section="connection")


app = typer.Typer(invoke_without_command=True, callback=configure)



import json
from typing import Any, Dict


def update_dict(
    target: Dict[str, str],
    source: Dict[str, str],
):
    for key, value in source.items():
        if value is not None:
            if key not in target:
                target[key] = value
            else:
                target[key] += value
    return target


def convert_str_to_type(value: str, type_str: str) -> Any:
    if type_str == "str":
        return value.strip()
    elif type_str == "bool":
        return value.lower() == "true"
    elif type_str == "int":
        return int(value)
    elif type_str == "float":
        return float(value)
    elif type_str.startswith("List"):
        return json.loads(value)
    elif type_str.startswith("Dict"):
        return json.loads(value)
    return value  # Default: Return as is



import os
from typing import Dict

import requests

import httpx
from rich import print

from promptmodel.utils.config_utils import read_config
from promptmodel.constants import ENDPOINT_URL
from promptmodel.utils.crypto import decrypt_message


class APIClient:
    """
    A class to represent an API request client.

    ...

    Methods
    -------
    get_headers():
        Generates headers for the API request.
    execute(method="GET", params=None, data=None, json=None, **kwargs):
        Executes the API request.
    """

    @classmethod
    def _get_headers(cls, use_cli_key: bool = True) -> Dict:
        """
        Reads, decrypts the api_key, and returns headers for API request.

        Returns
        -------
        dict
            a dictionary containing the Authorization header
        """
        config = read_config()
        if use_cli_key:
            if "connection" not in config:
                print(
                    "User not logged in. Please run [violet]prompt login[/violet] first."
                )
                exit()

            encrypted_key = (
                config["connection"]["encrypted_api_key"]
                if "encrypted_api_key" in config["connection"]
                else None
            )
            if encrypted_key is None:
                raise Exception("No API key found. Please run 'prompt login' first.")
            decrypted_key = decrypt_message(encrypted_key)
        else:
            decrypted_key = os.environ.get("PROMPTMODEL_API_KEY")
        headers = {"Authorization": f"Bearer {decrypted_key}"}
        return headers

    @classmethod
    def execute(
        cls,
        path: str,
        method="GET",
        params: Dict = None,
        data: Dict = None,
        json: Dict = None,
        ignore_auth_error: bool = False,
        use_cli_key: bool = True,
        **kwargs,
    ) -> requests.Response:
        """
        Executes the API request with the decrypted API key in the headers.

        Parameters
        ----------
        method : str, optional
            The HTTP method of the request (default is "GET")
        params : dict, optional
            The URL parameters to be sent with the request
        data : dict, optional
            The request body to be sent with the request
        json : dict, optional
            The JSON-encoded request body to be sent with the request
        ignore_auth_error: bool, optional
            Whether to ignore authentication errors (default is False)
        **kwargs : dict
            Additional arguments to pass to the requests.request function

        Returns
        -------
        requests.Response
            The response object returned by the requests library
        """
        url = f"{ENDPOINT_URL}{path}"
        headers = cls._get_headers(use_cli_key)
        try:
            response = requests.request(
                method,
                url,
                headers=headers,
                params=params,
                data=data,
                json=json,
                **kwargs,
            )
            if not response:
                print(f"[red]Error: {response}[/red]")
                exit()
            if response.status_code == 200:
                return response
            elif response.status_code == 403:
                if not ignore_auth_error:
                    print(
                        "[red]Authentication failed. Please run [violet][bold]prompt login[/bold][/violet] first.[/red]"
                    )
                    exit()
            else:
                print(f"[red]Error: {response}[/red]")
                exit()
        except requests.exceptions.ConnectionError:
            print("[red]Could not connect to the Promptmodel API.[/red]")
        except requests.exceptions.Timeout:
            print("[red]The request timed out.[/red]")


class AsyncAPIClient:
    """
    A class to represent an Async API request client.
    Used in Deployment stage.

    ...

    Methods
    -------
    get_headers():
        Generates headers for the API request.
    execute(method="GET", params=None, data=None, json=None, **kwargs):
        Executes the API request.
    """

    @classmethod
    async def _get_headers(cls, use_cli_key: bool = True) -> Dict:
        """
        Reads, decrypts the api_key, and returns headers for API request.

        Returns
        -------
        dict
            a dictionary containing the Authorization header
        """
        config = read_config()
        if use_cli_key:
            if "connection" not in config:
                print(
                    "User not logged in. Please run [violet]prompt login[/violet] first."
                )
                exit()

            encrypted_key = config["connection"]["encrypted_api_key"]
            if encrypted_key is None:
                raise Exception("No API key found. Please run 'prompt login' first.")
            decrypted_key = decrypt_message(encrypted_key)
        else:
            decrypted_key = os.environ.get("PROMPTMODEL_API_KEY")
            if decrypted_key is None:
                raise Exception(
                    "PROMPTMODEL_API_KEY was not found in the current environment."
                )
        headers = {"Authorization": f"Bearer {decrypted_key}"}
        return headers

    @classmethod
    async def execute(
        cls,
        path: str,
        method="GET",
        params: Dict = None,
        data: Dict = None,
        json: Dict = None,
        ignore_auth_error: bool = False,
        use_cli_key: bool = True,
        **kwargs,
    ) -> requests.Response:
        """
        Executes the API request with the decrypted API key in the headers.

        Parameters
        ----------
        method : str, optional
            The HTTP method of the request (default is "GET")
        params : dict, optional
            The URL parameters to be sent with the request
        data : dict, optional
            The request body to be sent with the request
        json : dict, optional
            The JSON-encoded request body to be sent with the request
        ignore_auth_error: bool, optional
            Whether to ignore authentication errors (default is False)
        **kwargs : dict
            Additional arguments to pass to the requests.request function

        Returns
        -------
        requests.Response
            The response object returned by the requests library
        """
        url = f"{ENDPOINT_URL}{path}"
        headers = await cls._get_headers(use_cli_key)
        try:
            async with httpx.AsyncClient(http2=True) as _client:
                response = await _client.request(
                    method,
                    url,
                    headers=headers,
                    params=params,
                    data=data,
                    json=json,
                    **kwargs,
                )
            if not response:
                print(f"[red]Error: {response}[/red]")
            if response.status_code == 200:
                return response
            elif response.status_code == 403:
                if not ignore_auth_error:
                    print("[red]Authentication failed.[/red]")
            else:
                print(f"[red]Error: {response}[/red]")

            return response
        except requests.exceptions.ConnectionError:
            print("[red]Could not connect to the Promptmodel API.[/red]")
        except requests.exceptions.Timeout:
            print("[red]The request timed out.[/red]")
        except Exception as exception:
            print(f"[red]Error: {exception}[/red]")



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.types.response import ChatModelConfig


async def echo_coroutine(*args, **kwargs):
    # print(args, kwargs)
    return args, kwargs


@pytest.fixture
def mock_fetch_chat_model():
    mock_fetch_chat_model = AsyncMock()
    mock_instruction = "You are a helpful assistant."
    mock_version_details = {
        "model": "gpt-4-1106-preview",
        "uuid": "testuuid",
        "version": 1,
    }
    mock_message_logs = [
        {
            "role": "system",
            "content": "You are a helpful assistant.",
            "session_uuid": "testuuid",
        },
        {"role": "user", "content": "Hello!", "session_uuid": "testuuid"},
        {
            "role": "assistant",
            "content": "Hello! How can I help you?",
            "session_uuid": "testuuid",
        },
    ]

    mock_fetch_chat_model.return_value = (
        mock_instruction,
        mock_version_details,
        mock_message_logs,
    )

    return mock_fetch_chat_model


@pytest.fixture
def mock_async_chat_log_to_cloud():
    mock_async_chat_log_to_cloud = AsyncMock()

    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_async_chat_log_to_cloud.return_value = mock_response

    return mock_async_chat_log_to_cloud


@pytest.fixture
def mock_async_make_session_cloud():
    mock_async_make_session_cloud = AsyncMock()

    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_async_make_session_cloud.return_value = mock_response

    return mock_async_make_session_cloud



import asyncio
from typing import Coroutine


def run_async_in_sync(coro: Coroutine):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:  # No running loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        result = loop.run_until_complete(coro)
        # loop.close()
        return result

    return loop.run_until_complete(coro)


def run_async_in_sync_threadsafe(coro: Coroutine, main_loop: asyncio.AbstractEventLoop):
    future = asyncio.run_coroutine_threadsafe(coro, main_loop)
    res = future.result()
    return res



# JSON Schema to pass to OpenAI
function_shemas = [
    {
        "name": "get_current_weather",
        "description": "Get the current weather in a given location",
        "parameters": {
            "type": "object",
            "properties": {
                "location": {
                    "type": "string",
                    "description": "The city and state, e.g. San Francisco, CA",
                },
                "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
            },
            "required": ["location"],
        },
    }
]



from typing import Any, Dict, List, Optional
from litellm import token_counter


def set_inputs_to_prompts(inputs: Dict[str, Any], prompts: List[Dict[str, str]]):
    messages = [
        {"content": prompt["content"].format(**inputs), "role": prompt["role"]}
        for prompt in prompts
    ]
    return messages


def num_tokens_for_messages(
    messages: List[Dict[str, str]], model: str = "gpt-3.5-turbo-0613"
) -> int:
    # tokens_per_message = 0
    # tokens_per_name = 0
    # if model.startswith("gpt-3.5-turbo"):
    #     tokens_per_message = 4
    #     tokens_per_name = -1

    # if model.startswith("gpt-4"):
    #     tokens_per_message = 3
    #     tokens_per_name = 1

    # if model.endswith("-0613") or model == "gpt-3.5-turbo-16k":
    #     tokens_per_message = 3
    #     tokens_per_name = 1
    # sum = 0
    processed_messages = [
        {**message, "function_call": str(message["function_call"])}
        if "function_call" in message
        else message
        for message in messages
    ]
    sum = token_counter(model=model, messages=processed_messages)
    # for message in messages:
    #     sum += tokens_per_message
    #     if "name" in message:
    #         sum += tokens_per_name
    return sum


def num_tokens_for_messages_for_each(
    messages: List[Dict[str, str]], model: str = "gpt-3.5-turbo-0613"
) -> List[int]:
    processed_messages = [
        {**message, "function_call": str(message["function_call"])}
        if "function_call" in message
        else message
        for message in messages
    ]
    processed_messages = [
        {**message, "tool_calls": str(message["tool_calls"])}
        if "tool_calls" in message
        else message
        for message in processed_messages
    ]
    return [
        token_counter(model=model, messages=[message]) for message in processed_messages
    ]


def num_tokens_from_functions_input(
    functions: Optional[List[Any]] = None, model="gpt-3.5-turbo-0613"
) -> int:
    """Return the number of tokens used by a list of functions."""
    if functions is None:
        return 0
    num_tokens = 0
    for function in functions:
        function_tokens = token_counter(model=model, text=function["name"])
        function_tokens += token_counter(model=model, text=function["description"])

        if "parameters" in function:
            parameters = function["parameters"]
            if "properties" in parameters:
                for properties_key in parameters["properties"]:
                    function_tokens += token_counter(model=model, text=properties_key)
                    v = parameters["properties"][properties_key]
                    for field in v:
                        if field == "type":
                            function_tokens += 2
                            function_tokens += token_counter(
                                model=model, text=v["type"]
                            )
                        elif field == "description":
                            function_tokens += 2
                            function_tokens += token_counter(
                                model=model, text=v["description"]
                            )
                        elif field == "enum":
                            function_tokens -= 3
                            for o in v["enum"]:
                                function_tokens += 3
                                function_tokens += token_counter(model=model, text=o)
                        else:
                            print(f"Warning: not supported field {field}")
                function_tokens += 11

        num_tokens += function_tokens

    num_tokens += 12
    return num_tokens


def num_tokens_from_function_call_output(
    function_call_output: Dict[str, str] = {}, model="gpt-3.5-turbo-0613"
) -> int:
    num_tokens = 1
    num_tokens += token_counter(model=model, text=function_call_output["name"])
    if "arguments" in function_call_output:
        num_tokens += token_counter(model=model, text=function_call_output["arguments"])
    return num_tokens



import typer
import os
from promptmodel.utils.config_utils import read_config, upsert_config
from promptmodel.database.orm import initialize_db


def fix():
    """Fix Bugs that can be occured in the promptmodel."""
    config = read_config()
    if "connection" in config:
        connection = config["connection"]
        if "initializing" in connection and connection["initializing"] == True:
            upsert_config({"initializing": False}, "connection")
        if "online" in connection and connection["online"] == True:
            upsert_config({"online": False}, "connection")
        if "reloading" in connection and connection["reloading"] == True:
            upsert_config({"reloading": False}, "connection")
    # delete .promptmodel/promptmodel.db
    # if .promptmodel/promptmodel.db exist, delete it
    if os.path.exists(".promptmodel/promptmodel.db"):
        os.remove(".promptmodel/promptmodel.db")
    # initialize_db()

    return


app = typer.Typer(invoke_without_command=True, callback=fix)



from enum import Enum


class InstanceType(str, Enum):
    ChatLog = "ChatLog"
    RunLog = "RunLog"
    ChatLogSession = "ChatLogSession"


class LocalTask(str, Enum):
    RUN_PROMPT_MODEL = "RUN_PROMPT_MODEL"
    RUN_CHAT_MODEL = "RUN_CHAT_MODEL"

    LIST_CODE_CHAT_MODELS = "LIST_CHAT_MODELS"
    LIST_CODE_PROMPT_MODELS = "LIST_PROMPT_MODELS"
    LIST_CODE_FUNCTIONS = "LIST_FUNCTIONS"


class LocalTaskErrorType(str, Enum):
    NO_FUNCTION_NAMED_ERROR = "NO_FUNCTION_NAMED_ERROR"  # no DB update is needed
    FUNCTION_CALL_FAILED_ERROR = "FUNCTION_CALL_FAILED_ERROR"  # create FunctionModelVersion, create Prompt, create RunLog
    PARSING_FAILED_ERROR = "PARSING_FAILED_ERROR"  # create FunctionModelVersion, create Prompt, create RunLog

    SERVICE_ERROR = "SERVICE_ERROR"  # no DB update is needed


class ServerTask(str, Enum):
    UPDATE_RESULT_RUN = "UPDATE_RESULT_RUN"
    UPDATE_RESULT_CHAT_RUN = "UPDATE_RESULT_CHAT_RUN"

    SYNC_CODE = "SYNC_CODE"


class Role(str, Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"
    FUNCTION = "function"


class ParsingType(str, Enum):
    COLON = "colon"
    SQUARE_BRACKET = "square_bracket"
    DOUBLE_SQUARE_BRACKET = "double_square_bracket"
    HTML = "html"


class ParsingPattern(dict, Enum):
    COLON = {
        "start": r"(\w+)\s+type=([\w,\s\[\]]+): \n",
        "start_fstring": "{key}: \n",
        "end_fstring": None,
        "whole": r"(.*?): (.*?)\n",
        "start_token": None,
        "end_token": None,
    }
    SQUARE_BRACKET = {
        "start": r"\[(\w+)\s+type=([\w,\s\[\]]+)\]",
        "start_fstring": "[{key} type={type}]",
        "end_fstring": "[/{key}]",
        "whole": r"\[(\w+)\s+type=([\w,\s\[\]]+)\](.*?)\[/\1\]",
        "start_token": r"[",
        "end_token": r"]",
    }
    DOUBLE_SQUARE_BRACKET = {
        "start": r"\[\[(\w+)\s+type=([\w,\s\[\]]+)\]\]",
        "start_fstring": "[[{key} type={type}]]",
        "end_fstring": "[[/{key}]]",
        "whole": r"\[\[(\w+)\s+type=([\w,\s\[\]]+)\]\](.*?)\[\[/\1\]\]",
        "start_token": r"[",
        "end_token": r"]",
    }
    HTML = {
        "start": r"<(\w+)\s+type=([\w,\s\[\]]+)>",
        "start_fstring": "<{key} type={type}>",
        "end_fstring": "</{key}>",
        "whole": r"<(\w+)\s+type=([\w,\s\[\]]+)>(.*?)</\1>",  # also captures type
        "start_token": r"<",
        "end_token": r">",
    }


def get_pattern_by_type(parsing_type_value):
    return ParsingPattern[ParsingType(parsing_type_value).name].value



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any
from litellm import ModelResponse

from promptmodel.llms.llm import LLM
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.types.response import LLMResponse, LLMStreamResponse



import os

from .models import (
    DeployedFunctionModel,
    DeployedFunctionModelVersion,
    DeployedPrompt,
)
from .config import db


def initialize_db():
    if not os.path.exists("./.promptmodel"):
        os.mkdir("./.promptmodel")
    # Check if db connection exists
    if db.is_closed():
        db.connect()
        with db.atomic():
            if not DeployedFunctionModel.table_exists():
                db.create_tables(
                    [
                        # FunctionModel,
                        # FunctionModelVersion,
                        # Prompt,
                        # RunLog,
                        # SampleInputs,
                        DeployedFunctionModel,
                        DeployedFunctionModelVersion,
                        DeployedPrompt,
                        # ChatModel,
                        # ChatModelVersion,
                        # ChatLogSession,
                        # ChatLog,
                    ]
                )
        db.close()
